# sage_setup: distribution = sagemath-combinat
r"""
Word classes

AUTHORS:

- Arnaud Bergeron
- Amy Glen
- Sébastien Labbé
- Franco Saliola
"""
#*****************************************************************************
#       Copyright (C) 2008 Arnaud Bergeron <abergeron@gmail.com>,
#                          Amy Glen <amy.glen@gmail.com>,
#                          Sébastien Labbé <slabqc@gmail.com>,
#                          Franco Saliola <saliola@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************
from sage.misc.lazy_import import lazy_import
from sage.combinat.words.word_char import WordDatatype_char
from sage.combinat.words.abstract_word import Word_class
from sage.combinat.words.finite_word import FiniteWord_class
from sage.combinat.words.infinite_word import InfiniteWord_class
from .word_datatypes import (WordDatatype_str,
                            WordDatatype_list,
                            WordDatatype_tuple)
from .word_infinite_datatypes import (
                            WordDatatype_iter_with_caching,
                            WordDatatype_iter,
                            WordDatatype_callable_with_caching,
                            WordDatatype_callable)
from .morphic import WordDatatype_morphic

lazy_import('sage.monoids.free_monoid_element', 'FreeMonoidElement')

# TODO. Word needs to be replaced by Word. Consider renaming
# Word_class to Word and imbedding Word as its __call__ method.


def Word(data=None, alphabet=None, length=None, datatype=None, caching=True, RSK_data=None):
    r"""
    Construct a word.

    INPUT:

    - ``data`` -- (default: ``None``) list, string, tuple, iterator, free
      monoid element, ``None`` (shorthand for ``[]``), or a callable defined
      on ``[0,1,...,length]``

    - ``alphabet`` -- any argument accepted by Words

    - ``length`` -- (default: ``None``) this is dependent on the type of data.
      It is ignored for words defined by lists, strings, tuples,
      etc., because they have a naturally defined length.
      For callables, this defines the domain of definition,
      which is assumed to be ``[0, 1, 2, ..., length-1]``.
      For iterators: Infinity if you know the iterator will not
      terminate (default); ``'unknown'`` if you do not know whether the
      iterator terminates; ``'finite'`` if you know that the iterator
      terminates, but do not know the length.

    - ``datatype`` -- (default: ``None``) ``None``, ``'list'``, ``'str'``,
      ``'tuple'``, ``'iter'``, ``'callable'``; if ``None``, then the function
      tries to guess this from the data

    - ``caching`` -- boolean (default: ``True``); whether to
      keep a cache of the letters computed by an iterator or callable

    - ``RSK_data`` -- (default: ``None``) semistandard and a
      standard Young tableau to run the inverse RSK bijection on

    .. NOTE::

       Be careful when defining words using callables and iterators. It
       appears that islice does not pickle correctly causing various errors
       when reloading. Also, most iterators do not support copying and
       should not support pickling by extension.

    EXAMPLES:

    Empty word::

        sage: Word()
        word:

    Word with string::

        sage: Word("abbabaab")
        word: abbabaab

    Word with string constructed from other types::

        sage: Word([0,1,1,0,1,0,0,1], datatype='str')
        word: 01101001
        sage: Word((0,1,1,0,1,0,0,1), datatype='str')
        word: 01101001

    Word with list::

        sage: Word([0,1,1,0,1,0,0,1])
        word: 01101001

    Word with list constructed from other types::

        sage: Word("01101001", datatype='list')
        word: 01101001
        sage: Word((0,1,1,0,1,0,0,1), datatype='list')
        word: 01101001

    Word with tuple::

        sage: Word((0,1,1,0,1,0,0,1))
        word: 01101001

    Word with tuple constructed from other types::

        sage: Word([0,1,1,0,1,0,0,1], datatype='tuple')
        word: 01101001
        sage: Word("01101001", datatype='str')
        word: 01101001

    Word with iterator::

        sage: from itertools import count
        sage: Word(count())
        word: 0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,...
        sage: Word(iter("abbabaab")) # iterators default to infinite words
        word: abbabaab
        sage: Word(iter("abbabaab"), length='unknown')
        word: abbabaab
        sage: Word(iter("abbabaab"), length='finite')
        word: abbabaab

    Word with function (a 'callable')::

        sage: f = lambda n : add(Integer(n).digits(2)) % 2
        sage: Word(f)
        word: 0110100110010110100101100110100110010110...
        sage: Word(f, length=8)
        word: 01101001

    Word over a string with a parent::

        sage: w = Word("abbabaab", alphabet='abc'); w
        word: abbabaab
        sage: w.parent()
        Finite words over {'a', 'b', 'c'}

    Word from a free monoid element::

        sage: M.<x,y,z> = FreeMonoid(3)
        sage: Word(x^3*y*x*z^2*x)
        word: xxxyxzzx

    The default parent is the combinatorial class of all words::

        sage: w = Word("abbabaab"); w
        word: abbabaab
        sage: w.parent()
        Finite words over Set of Python objects of class 'object'

    We can also input a semistandard tableau and a standard tableau to
    obtain a word from the inverse RSK algorithm using the
    ``RSK_data`` option::

        sage: p = Tableau([[1,2,2],[3]]); q = Tableau([[1,2,4],[3]])
        sage: Word(RSK_data=[p, q])
        word: 1322

    TESTS::

        sage: Word(5)
        Traceback (most recent call last):
        ...
        ValueError: cannot guess a datatype from data (=5); please specify one

    ::

        sage: W = Words()
        sage: w  = W('abc')
        sage: w is W(w)
        True
        sage: w is Word(w, alphabet='abc')
        False
    """
    if isinstance(data, FreeMonoidElement):
        return data.to_word(alphabet)

    if RSK_data is not None:
        #if a list of a semistandard and a standard tableau or a pair of lists
        from sage.combinat.tableau import Tableau
        if isinstance(RSK_data, (tuple, list)) and len(RSK_data) == 2 and \
            all(isinstance(x, Tableau) for x in RSK_data):
            from sage.combinat.rsk import RSK_inverse
            return RSK_inverse(*RSK_data, output='word')
        elif isinstance(RSK_data, (tuple, list)) and len(RSK_data) == 2 and \
            all(isinstance(x, (list, tuple)) for x in RSK_data):
            from sage.combinat.rsk import RSK_inverse
            P,Q = map(Tableau, RSK_data)
            return RSK_inverse(P, Q, 'word')
        raise ValueError("Invalid input. Must be a pair of tableaux")

    # Create the parent object
    from .words import Words
    parent = Words(alphabet)

    return parent(data=data, length=length, datatype=datatype, caching=caching)

#######################################################################
#                                                                     #
#                    Concrete word classes                            #
#                                                                     #
#######################################################################

##### Finite Words #####


class FiniteWord_char(WordDatatype_char, FiniteWord_class):
    r"""
    Finite word represented by an array ``unsigned char *`` (i.e. integers
    between 0 and 255).

    For any word ``w``, type ``w.<TAB>`` to see the functions that can be applied
    to ``w``.

    EXAMPLES::

        sage: W = Words(range(20))

        sage: w = W(list(range(1, 10)) * 2)
        sage: type(w)
        <class 'sage.combinat.words.word.FiniteWord_char'>
        sage: w
        word: 123456789123456789

        sage: w.is_palindrome()
        False
        sage: (w*w[::-1]).is_palindrome()
        True
        sage: (w[:-1:]*w[::-1]).is_palindrome()
        True

        sage: w.is_lyndon()
        False
        sage: W(list(range(10)) + [10, 10]).is_lyndon()
        True

        sage: w.is_square_free()
        False
        sage: w[:-1].is_square_free()
        True

        sage: u = W([randint(0,10) for i in range(10)])
        sage: (u*u).is_square()
        True
        sage: (u*u*u).is_cube()
        True

        sage: len(w.factor_set())
        127
        sage: w.rauzy_graph(5)                                                          # needs sage.graphs
        Looped digraph on 9 vertices

        sage: u = W([1,2,3])
        sage: w.first_occurrence(u)
        0
        sage: w.first_occurrence(u, start=1)
        9

    TESTS::

        sage: W = Words([0,1,2])
        sage: w = W([0,1,1,0])
        sage: w == loads(dumps(w))
        True
    """
    pass


class FiniteWord_list(WordDatatype_list, FiniteWord_class):
    r"""
    Finite word represented by a Python list.

    For any word `w`, type ``w.`` and hit :kbd:`Tab` key to see the list of
    functions defined on `w`.

    EXAMPLES::

        sage: w = Word(range(10))
        sage: w.iterated_right_palindromic_closure()
        word: 0102010301020104010201030102010501020103...

    TESTS::

        sage: w = Word([0,1,1,0])
        sage: w == loads(dumps(w))
        True
    """
    pass


class FiniteWord_str(WordDatatype_str, FiniteWord_class):
    r"""
    Finite word represented by a Python str.

    For such word `w`, type ``w.`` and hit :kbd:`Tab` key to see the list of
    functions defined on `w`.

    EXAMPLES::

        sage: w = Word('abcdef')
        sage: w.is_square()
        False

    TESTS::

        sage: w = Word('abba')
        sage: w == loads(dumps(w))
        True
    """
    pass


class FiniteWord_tuple(WordDatatype_tuple, FiniteWord_class):
    r"""
    Finite word represented by a Python tuple.

    For such word `w`, type ``w.`` and hit :kbd:`Tab` key to see the list of
    functions defined on `w`.

    EXAMPLES::

        sage: w = Word(())
        sage: w.is_empty()
        True

    TESTS::

        sage: w = Word((0,1,1,0))
        sage: w == loads(dumps(w))
        True
    """
    pass


class FiniteWord_iter_with_caching(WordDatatype_iter_with_caching, FiniteWord_class):
    r"""
    Finite word represented by an iterator (with caching).

    For such word `w`, type ``w.`` and hit :kbd:`Tab` key to see the list of
    functions defined on `w`.

    EXAMPLES::

        sage: w = Word(iter('abcdef'))
        sage: w.conjugate(2)
        word: cdefab

    TESTS::

        sage: w = Word(iter(range(10)))
        sage: type(w)
        <class 'sage.combinat.words.word.FiniteWord_iter_with_caching'>
        sage: z = loads(dumps(w))
        sage: w == z
        True
        sage: type(z)
        <class 'sage.combinat.words.word.FiniteWord_list'>
    """
    pass


class FiniteWord_iter(WordDatatype_iter, FiniteWord_class):
    r"""
    Finite word represented by an iterator.

    For such word `w`, type  ``w.`` and hit :kbd:`Tab` key to see the list of
    functions defined on `w`.

    EXAMPLES::

        sage: w = Word(iter(range(10)), caching=False)
        sage: w
        word: 0123456789
        sage: w.finite_differences()
        word: 111111111

    TESTS::

        sage: w = Word(iter(range(10)), caching=False)
        sage: type(w)
        <class 'sage.combinat.words.word.FiniteWord_iter'>
        sage: z = loads(dumps(w))
        sage: w == z
        True
        sage: type(z)
        <class 'sage.combinat.words.word.FiniteWord_list'>
    """
    pass


class FiniteWord_callable_with_caching(WordDatatype_callable_with_caching, FiniteWord_class):
    r"""
    Finite word represented by a callable (with caching).

    For such word `w`, type ``w.`` and hit :kbd:`Tab` key to see the list of
    functions defined on `w`.

    EXAMPLES::

        sage: f = lambda n : n % 3
        sage: w = Word(f, length=32)
        sage: w
        word: 01201201201201201201201201201201
        sage: w.border()
        word: 01201201201201201201201201201

    TESTS::

        sage: w = Word(lambda n:n, length=10)
        sage: type(w)
        <class 'sage.combinat.words.word.FiniteWord_callable_with_caching'>
        sage: z = loads(dumps(w))
        sage: w == z
        True
        sage: type(z)
        <class 'sage.combinat.words.word.FiniteWord_callable_with_caching'>

    Pickle also works for concatenation of words::

        sage: w = Word(range(10)) * Word('abcdef')
        sage: type(w)
        <class 'sage.combinat.words.word.FiniteWord_callable_with_caching'>
        sage: z = loads(dumps(w))
        sage: w == z
        True
        sage: type(z)
        <class 'sage.combinat.words.word.FiniteWord_list'>

    Pickle also works for power of words::

        sage: w = Word(range(10)) ^ 2
        sage: type(w)
        <class 'sage.combinat.words.word.FiniteWord_callable_with_caching'>
        sage: z = loads(dumps(w))
        sage: w == z
        True
        sage: type(z)
        <class 'sage.combinat.words.word.FiniteWord_list'>
    """
    pass


class FiniteWord_callable(WordDatatype_callable, FiniteWord_class):
    r"""
    Finite word represented by a callable.

    For such word `w`, type ``w.`` and hit :kbd:`Tab` key to see the list of
    functions defined on `w`.

    EXAMPLES::

        sage: f = lambda n : 3 if n > 8 else 6
        sage: w = Word(f, length=30, caching=False)
        sage: w
        word: 666666666333333333333333333333
        sage: w.is_symmetric()
        True

    TESTS::

        sage: w = Word(lambda n:n, length=10, caching=False)
        sage: type(w)
        <class 'sage.combinat.words.word.FiniteWord_callable'>
        sage: z = loads(dumps(w))
        sage: w == z
        True
        sage: type(z)
        <class 'sage.combinat.words.word.FiniteWord_callable'>
    """
    pass


##### Infinite Words #####

class InfiniteWord_iter_with_caching(WordDatatype_iter_with_caching, InfiniteWord_class):
    r"""
    Infinite word represented by an iterable (with caching).

    For such word `w`, type ``w.`` and hit :kbd:`Tab` key to see the list of
    functions defined on `w`.

    Infinite words behave like a Python list : they can be sliced using
    square braquets to define for example a prefix or a factor.

    EXAMPLES::

        sage: from itertools import cycle
        sage: w = Word(cycle([9,8,4]))
        sage: w
        word: 9849849849849849849849849849849849849849...
        sage: prefix = w[:23]
        sage: prefix
        word: 98498498498498498498498
        sage: prefix.minimal_period()
        3

    TESTS::

        sage: from itertools import count
        sage: w = Word(count())
        sage: type(w)
        <class 'sage.combinat.words.word.Word_iter_with_caching'>

    Pickle is not supported for infinite word defined by an iterator::

        sage: dumps(w)
        Traceback (most recent call last):
        ...
        TypeError: can...t...pickle...generator...object...
    """
    pass


class InfiniteWord_iter(WordDatatype_iter, InfiniteWord_class):
    r"""
    Infinite word represented by an iterable.

    For such word `w`, type ``w.`` and hit :kbd:`Tab` key to see the list of
    functions defined on `w`.

    Infinite words behave like a Python list : they can be sliced using
    square braquets to define for example a prefix or a factor.

    EXAMPLES::

        sage: from itertools import chain, cycle
        sage: w = Word(chain('letsgo', cycle('forever')), caching=False)
        sage: w
        word: letsgoforeverforeverforeverforeverforeve...
        sage: prefix = w[:100]
        sage: prefix
        word: letsgoforeverforeverforeverforeverforeve...
        sage: prefix.is_lyndon()
        False

    TESTS::

        sage: from itertools import count
        sage: w = Word(count(), caching=False)
        sage: type(w)
        <class 'sage.combinat.words.word.Word_iter'>

    Pickle is not supported for infinite word defined by an iterator::

        sage: dumps(w)
        Traceback (most recent call last):
        ...
        TypeError: can...t...pickle...generator...object...
    """
    pass


class InfiniteWord_callable_with_caching(WordDatatype_callable_with_caching, InfiniteWord_class):
    r"""
    Infinite word represented by a callable (with caching).

    For such word `w`, type ``w.`` and hit :kbd:`Tab` key to see the list of
    functions defined on `w`.

    Infinite words behave like a Python list : they can be sliced using
    square braquets to define for example a prefix or a factor.

    EXAMPLES::

        sage: w = Word(lambda n:n)
        sage: factor = w[4:13]
        sage: factor
        word: 4,5,6,7,8,9,10,11,12

    TESTS::

        sage: w = Word(lambda n:n)
        sage: type(w)
        <class 'sage.combinat.words.word.InfiniteWord_callable_with_caching'>
        sage: z = loads(dumps(w))
        sage: z
        word: 0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,...
        sage: type(z)
        <class 'sage.combinat.words.word.InfiniteWord_callable_with_caching'>
    """
    pass


class InfiniteWord_callable(WordDatatype_callable, InfiniteWord_class):
    r"""
    Infinite word represented by a callable.

    For such word `w`, type ``w.`` and hit :kbd:`Tab` key to see the list of
    functions defined on `w`.

    Infinite words behave like a Python list : they can be sliced using
    square braquets to define for example a prefix or a factor.

    EXAMPLES::

        sage: w = Word(lambda n:n, caching=False)
        sage: w
        word: 0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,...
        sage: w.iterated_right_palindromic_closure()
        word: 0102010301020104010201030102010501020103...

    TESTS::

        sage: w = Word(lambda n:n, caching=False)
        sage: type(w)
        <class 'sage.combinat.words.word.InfiniteWord_callable'>
        sage: z = loads(dumps(w))
        sage: z
        word: 0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,...
        sage: type(z)
        <class 'sage.combinat.words.word.InfiniteWord_callable'>
    """
    pass


##### Words of unknown length #####

class Word_iter_with_caching(WordDatatype_iter_with_caching, Word_class):
    r"""
    Word of unknown length (finite or infinite) represented by an
    iterable (with caching).

    For such word `w`, type ``w.`` and hit :kbd:`Tab` key to see the list of
    functions defined on `w`.

    Words behave like a Python list : they can be sliced using
    square braquets to define for example a prefix or a factor.

    EXAMPLES::

        sage: w = Word(iter([1,2,3]*1000), length='unknown')
        sage: w
        word: 1231231231231231231231231231231231231231...
        sage: w.finite_differences(mod=2)
        word: 1101101101101101101101101101101101101101...

    TESTS::

        sage: w = Word(iter('abcd'*100), length='unknown')
        sage: type(w)
        <class 'sage.combinat.words.word.Word_iter_with_caching'>
        sage: w
        word: abcdabcdabcdabcdabcdabcdabcdabcdabcdabcd...

    Pickle is not supported for word of unknown length defined by an iterator::

        sage: dumps(w)
        Traceback (most recent call last):
        ...
        TypeError: can...t...pickle...generator...object...
    """
    pass


class Word_iter(WordDatatype_iter, Word_class):
    r"""
    Word of unknown length (finite or infinite) represented by an
    iterable.

    For such word `w`, type ``w.`` and hit :kbd:`Tab` key to see the list of
    functions defined on `w`.

    Words behave like a Python list : they can be sliced using
    square braquets to define for example a prefix or a factor.

    EXAMPLES::

        sage: w = Word(iter([1,1,4,9]*1000), length='unknown', caching=False)
        sage: w
        word: 1149114911491149114911491149114911491149...
        sage: w.delta()
        word: 2112112112112112112112112112112112112112...

    TESTS::

        sage: w = Word(iter('abcd'*100), length='unknown', caching=False)
        sage: type(w)
        <class 'sage.combinat.words.word.Word_iter'>
        sage: w
        word: abcdabcdabcdabcdabcdabcdabcdabcdabcdabcd...

    Pickle is not supported for word of unknown length defined by an iterator::

        sage: dumps(w)
        Traceback (most recent call last):
        ...
        TypeError: can...t...pickle...generator...object...
    """
    pass


##### Morphic Words #####

class FiniteWord_morphic(WordDatatype_morphic, FiniteWord_class):
    r"""
    Finite morphic word.

    For such word `w`, type  ``w.`` and hit :kbd:`Tab` key to see the list of
    functions defined on `w`.

    EXAMPLES::

        sage: m = WordMorphism("a->ab,b->")
        sage: w = m.fixed_point("a")
        sage: w
        word: ab


    TESTS::

        sage: m = WordMorphism("a->ab,b->")
        sage: w = m.fixed_point("a")
        sage: type(w)
        <class 'sage.combinat.words.word.FiniteWord_morphic'>
        sage: loads(dumps(w))
        word: ab
    """
    pass


class InfiniteWord_morphic(WordDatatype_morphic, InfiniteWord_class):
    r"""
    Morphic word of infinite length.

    For such word `w`, type ``w.`` and hit :kbd:`Tab` key to see the list of
    functions defined on `w`.

    Infinite words behave like a Python list : they can be sliced using
    square braquets to define for example a prefix or a factor.

    EXAMPLES::

        sage: m = WordMorphism('a->ab,b->a')
        sage: w = m.fixed_point('a')
        sage: w
        word: abaababaabaababaababaabaababaabaababaaba...

    TESTS:

    Pickle is supported::

        sage: loads(dumps(w))
        word: abaababaabaababaababaabaababaabaababaaba...
    """
    pass



# sage_setup: distribution = sagemath-combinat
r"""
Decorated permutations

AUTHORS:

- Martin Rubey (2020): Initial version
"""
# ****************************************************************************
#       Copyright (C) 2020 Martin Rubey <martin.rubey at tuwien.ac.at>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful, but
#    WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.misc.inherit_comparison import InheritComparisonClasscallMetaclass
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.parent import Parent
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.arith.misc import factorial
from sage.rings.integer import Integer
from sage.combinat.permutation import Permutations
from sage.combinat.subset import Subsets
from sage.combinat.colored_permutations import SignedPermutations
from sage.structure.list_clone import ClonableArray


class DecoratedPermutation(ClonableArray,
        metaclass=InheritComparisonClasscallMetaclass):
    r"""
    A decorated permutation.

    A decorated permutation is a signed permutation where all
    non-fixed points have positive sign.
    """
    @staticmethod
    def __classcall_private__(cls, pi):
        """
        Create a decorated permutation.

        EXAMPLES::

            sage: DecoratedPermutation([2, 1, 3])
            [2, 1, 3]

            sage: DecoratedPermutation([2, 1, -3])
            [2, 1, -3]

        TESTS:

        Check that hashing and comparison works::

            sage: S = DecoratedPermutations(3)
            sage: elt1 = S([2, 1, -3])
            sage: elt2 = DecoratedPermutation([2, 1, -3])
            sage: elt1 == elt2
            True

            sage: elt1 == [2, 1, -3]
            False

            sage: elt2 = DecoratedPermutation([2, 1, 3])
            sage: elt1 != elt2
            True

            sage: hash(elt1)                                                    # random
            915443076393556996
        """
        pi = list(pi)
        return DecoratedPermutations(len(pi))(pi)

    def __init__(self, parent, pi, check=True):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: S = DecoratedPermutations(3)
            sage: elt = S([2, 1, -3])
            sage: TestSuite(elt).run()
        """
        ClonableArray.__init__(self, parent, pi, check=check)

    def check(self):
        """
        Check that ``self`` is a valid decorated permutation.

        EXAMPLES::

            sage: S = DecoratedPermutations(3)
            sage: elt = S([2, 1, -3])
            sage: elt.check()
            sage: elt = S([2, -1, 3])
            Traceback (most recent call last):
            ...
            ValueError: invalid decorated permutation
        """
        if self not in self.parent():
            raise ValueError("{} is not a decorated permutation".format(self))

    def size(self):
        """
        Return the size of the decorated permutation.

        EXAMPLES::

            sage: DecoratedPermutation([2, 1, -3]).size()
            3
        """
        return len(self)

    def to_signed_permutation(self):
        """
        Return ``self`` as a signed permutation.

        EXAMPLES::

            sage: DecoratedPermutation([2, 1, -3]).to_signed_permutation()
            [2, 1, -3]
        """
        return SignedPermutations(len(self))(list(self))


class DecoratedPermutations(UniqueRepresentation, Parent):
    r"""
    Class of all decorated permutations of `n`.

    A decorated permutation is a signed permutation where all
    non-fixed points have positive sign.

    INPUT:

    - ``n`` -- integer; the size of the decorated permutations

    EXAMPLES:

    This will create an instance to manipulate the decorated
    permutations of size 3::

        sage: S = DecoratedPermutations(3); S
        Decorated permutations of size 3
        sage: S.cardinality()
        16
    """

    def __init__(self, n):
        r"""
        Initialize ``self``.

        TESTS::

            sage: S = DecoratedPermutations(4)
            sage: TestSuite(S).run()
        """
        self._n = n
        Parent.__init__(self, category=FiniteEnumeratedSets())

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        TESTS::

            sage: DecoratedPermutations(4)
            Decorated permutations of size 4
        """
        return "Decorated permutations of size %s" % self._n

    def __contains__(self, pi):
        """
        Check if ``pi`` is in ``self``.

        TESTS::

            sage: S = DecoratedPermutations(3)
            sage: [2, 1, -3] in S
            True
            sage: [2, -1, 3] in S
            False
        """
        if isinstance(pi, DecoratedPermutation):
            return len(pi) == self._n

        values = list(pi)
        if len(values) != self._n:
            return False
        abs_values = [abs(v) for v in values]
        for i, (v, abs_v) in enumerate(zip(values, abs_values), 1):
            if i != abs_v and v < 0:
                return False
        return sorted(abs_values) == list(range(1, self._n + 1))

    def _element_constructor_(self, pi, check=True):
        """
        Construct an element of ``self``.

        EXAMPLES::

            sage: S = DecoratedPermutations(3)
            sage: elt = S([2, 1, -3])
            sage: elt.parent() is S
            True
        """
        if isinstance(pi, DecoratedPermutation):
            if pi.parent() is self:
                return pi
            raise ValueError("cannot convert between decorated permutations of different sizes")

        pi = tuple(pi)
        if check and pi not in self:
            raise ValueError('invalid decorated permutation')
        return self.element_class(self, pi)

    Element = DecoratedPermutation

    def _an_element_(self):
        """
        Return an element of ``self``.

        EXAMPLES::

            sage: S = DecoratedPermutations(3)
            sage: S._an_element_()
            [1, 2, 3]
        """
        return self.element_class(self, list(range(1, self._n + 1)))

    def cardinality(self):
        r"""
        Return the cardinality of ``self``.

        The number of decorated permutations of size `n` is equal to

        .. MATH::

            \sum_{k=0^n} \frac{n!}{k!}

        EXAMPLES::

            sage: [DecoratedPermutations(n).cardinality() for n in range(11)]
            [1, 2, 5, 16, 65, 326, 1957, 13700, 109601, 986410, 9864101]
        """
        return Integer(sum(factorial(self._n) // factorial(k)
                           for k in range(self._n + 1)))

    def __iter__(self):
        r"""
        Iterator on the decorated permutations of size `n`.

        TESTS::

            sage: S = DecoratedPermutations(2); S.list()
            [[1, 2], [-1, 2], [1, -2], [-1, -2], [2, 1]]
            sage: sum(1 for a in S)
            5
        """
        for sigma in Permutations(self._n):
            F = sigma.fixed_points()
            for X in Subsets(F):
                tau = list(sigma)
                for i in X:
                    tau[i - 1] = -tau[i - 1]
                yield DecoratedPermutations(self._n)(tau)



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.graphs sage.modules
"""
Direct sum of crystals
"""
# ****************************************************************************
#       Copyright (C) 2010 Anne Schilling <anne at math.ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
# ***************************************************************************

from sage.categories.category import Category
from sage.sets.disjoint_union_enumerated_sets import DisjointUnionEnumeratedSets
from sage.sets.family import Family
from sage.structure.element_wrapper import ElementWrapper
from sage.structure.element import get_coercion_model


class DirectSumOfCrystals(DisjointUnionEnumeratedSets):
    r"""
    Direct sum of crystals.

    Given a list of crystals `B_0, \ldots, B_k` of the same Cartan type,
    one can form the direct sum `B_0 \oplus \cdots \oplus B_k`.

    INPUT:

    - ``crystals`` -- list of crystals of the same Cartan type
    - ``keepkey`` -- boolean

    The option ``keepkey`` is by default set to ``False``, assuming
    that the crystals are all distinct. In this case the elements of
    the direct sum are just represented by the elements in the
    crystals `B_i`.  If the crystals are not all distinct, one should
    set the ``keepkey`` option to ``True``.  In this case, the
    elements of the direct sum are represented as tuples `(i, b)`
    where `b \in B_i`.

    EXAMPLES::

        sage: C = crystals.Letters(['A',2])
        sage: C1 = crystals.Tableaux(['A',2],shape=[1,1])
        sage: B = crystals.DirectSum([C,C1])
        sage: B.list()
        [1, 2, 3, [[1], [2]], [[1], [3]], [[2], [3]]]
        sage: [b.f(1) for b in B]
        [2, None, None, None, [[2], [3]], None]
        sage: B.module_generators
        (1, [[1], [2]])

    ::

        sage: B = crystals.DirectSum([C,C], keepkey=True)
        sage: B.list()
        [(0, 1), (0, 2), (0, 3), (1, 1), (1, 2), (1, 3)]
        sage: B.module_generators
        ((0, 1), (1, 1))
        sage: b = B( tuple([0,C(1)]) )
        sage: b
        (0, 1)
        sage: b.weight()
        (1, 0, 0)

    The following is required, because
    :class:`~sage.combinat.crystals.direct_sum.DirectSumOfCrystals`
    takes the same arguments as :class:`DisjointUnionEnumeratedSets`
    (which see for details).

    TESTS::

        sage: C = crystals.Letters(['A',2])
        sage: B = crystals.DirectSum([C,C], keepkey=True)
        sage: B
        Direct sum of the crystals Family (The crystal of letters for type ['A', 2], The crystal of letters for type ['A', 2])

        sage: TestSuite(C).run()
    """
    @staticmethod
    def __classcall_private__(cls, crystals, facade=True, keepkey=False, category=None):
        """
        Normalization of arguments; see :class:`UniqueRepresentation`.

        TESTS:

        We check that direct sum of crystals have unique representation::

            sage: B = crystals.Tableaux(['A',2], shape=[2,1])
            sage: C = crystals.Letters(['A',2])
            sage: D1 = crystals.DirectSum([B, C])
            sage: D2 = crystals.DirectSum((B, C))
            sage: D1 is D2
            True
            sage: D3 = crystals.DirectSum([B, C, C])
            sage: D4 = crystals.DirectSum([D1, C])
            sage: D3 is D4
            True
        """
        if not isinstance(facade, bool) or not isinstance(keepkey, bool):
            raise TypeError
        # Normalize the facade-keepkey by giving keepkey dominance
        facade = not keepkey

        # We expand out direct sums of crystals
        ret = []
        for x in Family(crystals):
            if isinstance(x, DirectSumOfCrystals):
                ret += list(x.crystals)
            else:
                ret.append(x)
        category = Category.meet([Category.join(c.categories()) for c in ret])
        return super().__classcall__(cls,
            Family(ret), facade=facade, keepkey=keepkey, category=category)

    def __init__(self, crystals, facade, keepkey, category, **options):
        """
        TESTS::

            sage: C = crystals.Letters(['A',2])
            sage: B = crystals.DirectSum([C,C], keepkey=True)
            sage: B
            Direct sum of the crystals Family (The crystal of letters for type ['A', 2], The crystal of letters for type ['A', 2])
            sage: B.cartan_type()
            ['A', 2]

            sage: from sage.combinat.crystals.direct_sum import DirectSumOfCrystals
            sage: isinstance(B, DirectSumOfCrystals)
            True
        """
        DisjointUnionEnumeratedSets.__init__(self, crystals, keepkey=keepkey,
                                             facade=facade, category=category)
        self.rename("Direct sum of the crystals {}".format(crystals))
        self._keepkey = keepkey
        self.crystals = crystals
        if len(crystals) == 0:
            raise ValueError("the direct sum is empty")
        else:
            assert all(crystal.cartan_type() == crystals[0].cartan_type() for crystal in crystals)
            self._cartan_type = crystals[0].cartan_type()
        if keepkey:
            self.module_generators = tuple([self((i, b))
                                            for i, B in enumerate(crystals)
                                            for b in B.module_generators])
        else:
            self.module_generators = sum((tuple(B.module_generators) for B in crystals), ())

    def weight_lattice_realization(self):
        r"""
        Return the weight lattice realization used to express weights.

        The weight lattice realization is the common parent which all
        weight lattice realizations of the crystals of ``self`` coerce
        into.

        EXAMPLES::

            sage: LaZ = RootSystem(['A',2,1]).weight_lattice(extended=True).fundamental_weights()
            sage: LaQ = RootSystem(['A',2,1]).weight_space(extended=True).fundamental_weights()
            sage: B = crystals.LSPaths(LaQ[1])
            sage: B.weight_lattice_realization()
            Extended weight space over the Rational Field of the Root system of type ['A', 2, 1]
            sage: C = crystals.AlcovePaths(LaZ[1])
            sage: C.weight_lattice_realization()
            Extended weight lattice of the Root system of type ['A', 2, 1]
            sage: D = crystals.DirectSum([B,C])
            sage: D.weight_lattice_realization()
            Extended weight space over the Rational Field of the Root system of type ['A', 2, 1]
        """
        cm = get_coercion_model()
        return cm.common_parent(*[crystal.weight_lattice_realization()
                                  for crystal in self.crystals])

    class Element(ElementWrapper):
        r"""
        A class for elements of direct sums of crystals.
        """

        def e(self, i):
            r"""
            Return the action of `e_i` on ``self``.

            EXAMPLES::

                sage: C = crystals.Letters(['A',2])
                sage: B = crystals.DirectSum([C,C], keepkey=True)
                sage: [[b, b.e(2)] for b in B]
                [[(0, 1), None], [(0, 2), None], [(0, 3), (0, 2)], [(1, 1), None], [(1, 2), None], [(1, 3), (1, 2)]]
            """
            v = self.value
            vn = v[1].e(i)
            if vn is None:
                return None
            else:
                return self.parent()(tuple([v[0],vn]))

        def f(self, i):
            r"""
            Return the action of `f_i` on ``self``.

            EXAMPLES::

                sage: C = crystals.Letters(['A',2])
                sage: B = crystals.DirectSum([C,C], keepkey=True)
                sage: [[b,b.f(1)] for b in B]
                [[(0, 1), (0, 2)], [(0, 2), None], [(0, 3), None], [(1, 1), (1, 2)], [(1, 2), None], [(1, 3), None]]
            """
            v = self.value
            vn = v[1].f(i)
            if vn is None:
                return None
            else:
                return self.parent()(tuple([v[0],vn]))

        def weight(self):
            r"""
            Return the weight of ``self``.

            EXAMPLES::

                sage: C = crystals.Letters(['A',2])
                sage: B = crystals.DirectSum([C,C], keepkey=True)
                sage: b = B( tuple([0,C(2)]) )
                sage: b
                (0, 2)
                sage: b.weight()
                (0, 1, 0)
            """
            return self.value[1].weight()

        def phi(self, i):
            r"""
            EXAMPLES::

                sage: C = crystals.Letters(['A',2])
                sage: B = crystals.DirectSum([C,C], keepkey=True)
                sage: b = B( tuple([0,C(2)]) )
                sage: b.phi(2)
                1
            """
            return self.value[1].phi(i)

        def epsilon(self, i):
            r"""
            EXAMPLES::

                sage: C = crystals.Letters(['A',2])
                sage: B = crystals.DirectSum([C,C], keepkey=True)
                sage: b = B( tuple([0,C(2)]) )
                sage: b.epsilon(2)
                0
            """
            return self.value[1].epsilon(i)



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.graphs sage.modules
r"""
Rigged configurations

AUTHORS:

- Travis Scrimshaw (2010-09-26): initial version
"""

# ****************************************************************************
#       Copyright (C) 2010-2012 Travis Scrimshaw <tscrim@ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import itertools

from sage.misc.cachefunc import cached_method
from sage.misc.lazy_attribute import lazy_attribute
from sage.structure.global_options import GlobalOptions
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.parent import Parent
from sage.combinat.misc import IterableFunctionCall
import sage.combinat.tableau as tableau
from sage.rings.rational_field import QQ
from sage.categories.loop_crystals import KirillovReshetikhinCrystals
from sage.combinat.root_system.cartan_type import CartanType
from sage.combinat.rigged_configurations.kleber_tree import KleberTree, VirtualKleberTree
from sage.combinat.rigged_configurations.rigged_configuration_element import (
     RiggedConfigurationElement, KRRCSimplyLacedElement, KRRCNonSimplyLacedElement,
     KRRCTypeA2DualElement)
from sage.combinat.rigged_configurations.rigged_partition import RiggedPartition


# Used in the KR crystals catalog so that there is a common interface
def KirillovReshetikhinCrystal(cartan_type, r, s):
    r"""
    Return the KR crystal `B^{r,s}` using
    :class:`rigged configurations <RiggedConfigurations>`.

    This is the rigged configuration `RC(B^{r,s})` or `RC(L)` with
    `L = (L_i^{(a)})` and `L_i^{(a)} = \delta_{a,r} \delta_{i,s}`.

    EXAMPLES::

        sage: K1 = crystals.kirillov_reshetikhin.RiggedConfigurations(['A',6,2], 2, 1)
        sage: K2 = crystals.kirillov_reshetikhin.LSPaths(['A',6,2], 2, 1)
        sage: K1.digraph().is_isomorphic(K2.digraph(), edge_labels=True)
        True

    TESTS:

    We explicitly import and check we get the same crystal::

        sage: from sage.combinat.rigged_configurations.rigged_configurations import KirillovReshetikhinCrystal
        sage: K1 = crystals.kirillov_reshetikhin.RiggedConfigurations(['A',6,2], 2, 1)
        sage: K1 is KirillovReshetikhinCrystal(['A',6,2], 2, 1)
        True
    """
    return RiggedConfigurations(cartan_type, [[r, s]])


# Note on implementation, this class is used for simply-laced types only
class RiggedConfigurations(UniqueRepresentation, Parent):
    r"""
    Rigged configurations as `U_q^{\prime}(\mathfrak{g})`-crystals.

    Let `\overline{I}` denote the classical index set associated to the Cartan
    type of the rigged configurations. A rigged configuration of multiplicity
    array `L_i^{(a)}` and dominant weight `\Lambda` is a sequence of partitions
    `\{ \nu^{(a)} \mid a \in \overline{I} \}` such that

    .. MATH::

        \sum_{\overline{I} \times \ZZ_{>0}} i m_i^{(a)} \alpha_a
        = \sum_{\overline{I} \times \ZZ_{>0}} i L_i^{(a)} \Lambda_a
        - \Lambda

    where `\alpha_a` is a simple root, `\Lambda_a` is a fundamental weight,
    and `m_i^{(a)}` is the number of rows of length `i` in the partition
    `\nu^{(a)}`.

    Each partition `\nu^{(a)}`, in the sequence also comes with a sequence of
    statistics `p_i^{(a)}` called *vacancy numbers* and a weakly decreasing
    sequence `J_i^{(a)}` of length `m_i^{(a)}` called *riggings*.
    Vacancy numbers are computed based upon the partitions and `L_i^{(a)}`,
    and the riggings must satisfy `\max J_i^{(a)} \leq p_i^{(a)}`. We call
    such a partition a *rigged partition*. For more, see
    [RigConBijection]_ [CrysStructSchilling06]_ [BijectionLRT]_.

    Rigged configurations form combinatorial objects first introduced by
    Kerov, Kirillov and Reshetikhin that arose from studies of statistical
    mechanical models using the Bethe Ansatz. They are sequences of rigged
    partitions. A rigged partition is a partition together with a label
    associated to each part that satisfy certain constraints. The labels
    are also called riggings.

    Rigged configurations exist for all affine Kac-Moody Lie algebras. See
    for example [HKOTT2002]_. In Sage they are specified by providing a Cartan
    type and a list of rectangular shapes `B`. The list of all (highest
    weight) rigged configurations for given `B` is computed via the (virtual)
    Kleber algorithm (see also
    :class:`~sage.combinat.rigged_configurations.kleber_tree.KleberTree` and
    :class:`~sage.combinat.rigged_configurations.kleber_tree.VirtualKleberTree`).

    Rigged configurations in simply-laced types all admit a classical crystal
    structure [CrysStructSchilling06]_. For non-simply-laced types, the
    crystal is given by using virtual rigged configurations [OSS03]_. The
    highest weight rigged configurations are those where all riggings are
    nonnegative. The list of all rigged configurations is computed from the
    highest weight ones using the crystal operators.

    Rigged configurations are conjecturally in bijection with
    :class:`~sage.combinat.rigged_configurations.tensor_product_kr_tableaux.TensorProductOfKirillovReshetikhinTableaux`
    of non-exceptional affine types where the list `B` corresponds to the
    tensor factors `B^{r,s}`. The bijection has been proven in types `A_n^{(1)}`
    and `D_n^{(1)}` and when the only nonzero entries of `L_i^{(a)}` are either
    only `L_1^{(a)}` or only `L_i^{(1)}` (corresponding to single columns or
    rows respectively) [RigConBijection]_, [BijectionLRT]_, [BijectionDn]_.

    KR crystals are implemented in Sage, see
    :func:`~sage.combinat.crystals.kirillov_reshetikhin.KirillovReshetikhinCrystal`,
    however, in the bijection with rigged configurations a different
    realization of the elements in the crystal are obtained, which are
    coined KR tableaux, see
    :class:`~sage.combinat.rigged_configurations.kr_tableaux.KirillovReshetikhinTableaux`.
    For more details see [OSS2011]_.

    .. NOTE::

        All non-simply-laced rigged configurations have not been proven to
        give rise to aligned virtual crystals (i.e. have the correct crystal
        structure or isomorphic as affine crystals to the tensor product of
        KR tableaux).

    INPUT:

    - ``cartan_type`` -- a Cartan type

    - ``B`` -- list of positive integer tuples `(r,s)` corresponding to the
      tensor factors in the bijection with tensor product of
      Kirillov-Reshetikhin tableaux or equivalently the sequence of width `s`
      and height `r` rectangles

    REFERENCES:

    .. [HKOTT2002] \G. Hatayama, A. Kuniba, M. Okado, T. Takagi, Z. Tsuboi.
       Paths, Crystals and Fermionic Formulae.
       Prog. Math. Phys. **23** (2002) Pages 205-272.

    .. [CrysStructSchilling06] Anne Schilling.
       Crystal structure on rigged configurations.
       International Mathematics Research Notices.
       Volume 2006. (2006) Article ID 97376. Pages 1-27.

    .. [RigConBijection] Masato Okado, Anne Schilling, Mark Shimozono.
       A crystal to rigged configuration bijection for non-exceptional affine
       algebras.
       Algebraic Combinatorics and Quantum Groups.
       Edited by N. Jing. World Scientific. (2003) Pages 85-124.

    .. [BijectionDn] Anne Schilling.
       A bijection between type `D_n^{(1)}` crystals and rigged configurations.
       J. Algebra. **285** (2005) 292-334

    .. [BijectionLRT] Anatol N. Kirillov, Anne Schilling, Mark Shimozono.
       A bijection between Littlewood-Richardson tableaux and rigged
       configurations.
       Selecta Mathematica (N.S.). **8** (2002) Pages 67-135.
       (:mathscinet:`MR1890195`).

    EXAMPLES::

        sage: RC = RiggedConfigurations(['A', 3, 1], [[3, 2], [1, 2], [1, 1]])
        sage: RC
        Rigged configurations of type ['A', 3, 1] and factor(s) ((3, 2), (1, 2), (1, 1))

        sage: RC = RiggedConfigurations(['A', 3, 1], [[2,1]]); RC
        Rigged configurations of type ['A', 3, 1] and factor(s) ((2, 1),)
        sage: RC.cardinality()
        6
        sage: len(RC.list()) == RC.cardinality()
        True
        sage: RC.list() # random
        [
        <BLANKLINE>
                                                 0[ ]0
        (/)  (/)      (/)      -1[ ]-1  -1[ ]-1
                                                 -1[ ]-1
        (/)  -1[ ]-1  0[ ]0    0[ ]0    1[ ]1    -1[ ]-1
        <BLANKLINE>
        (/)  (/)      -1[ ]-1  (/)      -1[ ]-1  0[ ]0
           ,        ,        ,        ,        ,
        ]

    A rigged configuration element with all riggings equal to the vacancy
    numbers can be created as follows::

        sage: RC = RiggedConfigurations(['A', 3, 1], [[3,2], [2,1], [1,1], [1,1]]); RC
        Rigged configurations of type ['A', 3, 1] and factor(s) ((3, 2), (2, 1), (1, 1), (1, 1))
        sage: elt = RC(partition_list=[[1],[],[]]); elt
        <BLANKLINE>
        0[ ]0
        <BLANKLINE>
        (/)
        <BLANKLINE>
        (/)
        <BLANKLINE>

    If on the other hand we also want to specify the riggings, this can be
    achieved as follows::

        sage: RC = RiggedConfigurations(['A', 3, 1], [[3, 2], [1, 2], [1, 1]])
        sage: RC(partition_list=[[2],[2],[2]])
        <BLANKLINE>
        1[ ][ ]1
        <BLANKLINE>
        0[ ][ ]0
        <BLANKLINE>
        0[ ][ ]0
        sage: RC(partition_list=[[2],[2],[2]], rigging_list=[[0],[0],[0]])
        <BLANKLINE>
        1[ ][ ]0
        <BLANKLINE>
        0[ ][ ]0
        <BLANKLINE>
        0[ ][ ]0

    A larger example::

        sage: RC = RiggedConfigurations(['D', 7, 1], [[3,3],[5,2],[4,3],[2,3],[4,4],[3,1],[1,4],[2,2]])
        sage: elt = RC(partition_list=[[2],[3,2,1],[2,2,1,1],[2,2,1,1,1,1],[3,2,1,1,1,1],[2,1,1],[2,2]],
        ....:          rigging_list=[[2],[1,0,0],[4,1,2,1],[1,0,0,0,0,0],[0,1,0,0,0,0],[0,0,0],[0,0]])
        sage: elt
        <BLANKLINE>
        3[ ][ ]2
        <BLANKLINE>
        1[ ][ ][ ]1
        2[ ][ ]0
        1[ ]0
        <BLANKLINE>
        4[ ][ ]4
        4[ ][ ]1
        3[ ]2
        3[ ]1
        <BLANKLINE>
        2[ ][ ]1
        2[ ][ ]0
        0[ ]0
        0[ ]0
        0[ ]0
        0[ ]0
        <BLANKLINE>
        0[ ][ ][ ]0
        2[ ][ ]1
        0[ ]0
        0[ ]0
        0[ ]0
        0[ ]0
        <BLANKLINE>
        0[ ][ ]0
        0[ ]0
        0[ ]0
        <BLANKLINE>
        0[ ][ ]0
        0[ ][ ]0
        <BLANKLINE>

    To obtain the KR tableau under the bijection between rigged configurations
    and KR tableaux, we can type the following. This example was checked
    against Reiho Sakamoto's Mathematica program on rigged configurations::

        sage: output = elt.to_tensor_product_of_kirillov_reshetikhin_tableaux(); output
        [[1, 1, 1], [2, 3, 3], [3, 4, -5]] (X) [[1, 1], [2, 2], [3, 3], [5, -6], [6, -5]] (X)
        [[1, 1, 2], [2, 2, 3], [3, 3, 7], [4, 4, -7]] (X) [[1, 1, 1], [2, 2, 2]] (X)
        [[1, 1, 1, 3], [2, 2, 3, 4], [3, 3, 4, 5], [4, 4, 5, 6]] (X) [[1], [2], [3]] (X) [[1, 1, 1, 1]] (X) [[1, 1], [2, 2]]
        sage: elt.to_tensor_product_of_kirillov_reshetikhin_tableaux().to_rigged_configuration() == elt
        True
        sage: output.to_rigged_configuration().to_tensor_product_of_kirillov_reshetikhin_tableaux() == output
        True

    We can also convert between rigged configurations and tensor products of
    KR crystals::

        sage: RC = RiggedConfigurations(['D', 4, 1], [[2, 1]])
        sage: elt = RC(partition_list=[[1],[1,1],[1],[1]])
        sage: tp_krc = elt.to_tensor_product_of_kirillov_reshetikhin_crystals(); tp_krc
        [[]]
        sage: ret = RC(tp_krc)
        sage: ret == elt
        True

    ::

        sage: RC = RiggedConfigurations(['D', 4, 1], [[4,1], [3,3]])
        sage: KR1 = crystals.KirillovReshetikhin(['D', 4, 1], 4, 1)
        sage: KR2 = crystals.KirillovReshetikhin(['D', 4, 1], 3, 3)
        sage: T = crystals.TensorProduct(KR1, KR2)
        sage: t = T[1]; t
        [[++++, []], [+++-, [[1], [2], [4], [-4]]]]
        sage: ret = RC(t)
        sage: ret.to_tensor_product_of_kirillov_reshetikhin_crystals()
        [[++++, []], [+++-, [[1], [2], [4], [-4]]]]

    TESTS::

        sage: RC = RiggedConfigurations(['A', 3, 1], [[3,2], [2,1], [1,1], [1,1]])
        sage: len(RC.module_generators)
        17
        sage: RC = RiggedConfigurations(['D', 4, 1], [[1, 1]])
        sage: RC.cardinality()
        8

        sage: RC = RiggedConfigurations(['D', 4, 1], [[2, 1]])
        sage: c = RC.cardinality(); c
        29
        sage: K = crystals.KirillovReshetikhin(['D',4,1],2,1)
        sage: K.cardinality() == c
        True
    """
    @staticmethod
    def __classcall_private__(cls, cartan_type, B):
        r"""
        Normalize the input arguments to ensure unique representation.

        EXAMPLES::

            sage: RC1 = RiggedConfigurations(CartanType(['A',3,1]), [[2,2]])
            sage: RC2 = RiggedConfigurations(['A',3,1], [(2,2)])
            sage: RC3 = RiggedConfigurations(['A',3,1], ((2,2),))
            sage: RC2 is RC1, RC3 is RC1
            (True, True)
        """
        cartan_type = CartanType(cartan_type)
        if not cartan_type.is_affine():
            raise ValueError("The Cartan type must be affine")

        # Standardize B input into a tuple of tuples
        B = tuple(tuple(factor) for factor in B)
        if not B:
            raise ValueError("must contain at least one factor")

        if cartan_type.type() == 'BC': # Type `A_{2n}^{(2)}`
            return RCTypeA2Even(cartan_type, B)
        if cartan_type.dual().type() == 'BC': # Type 'A_{2n}^{(2)\dagger`
            return RCTypeA2Dual(cartan_type, B)
        # We check the classical type to account for A^{(1)}_1 which is not
        #    a virtual rigged configuration.
        if not cartan_type.classical().is_simply_laced():
            return RCNonSimplyLaced(cartan_type, B)

        return super().__classcall__(cls, cartan_type, B)

    def __init__(self, cartan_type, B):
        r"""
        Initialize the RiggedConfigurations class.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['A', 3, 1], [[3,1], [1,2]])
            sage: TestSuite(RC).run() # long time
            sage: RC = RiggedConfigurations(['A',1,1], [[1,1], [1,1]])
            sage: TestSuite(RC).run()
            sage: RC = RiggedConfigurations(['A',2,1], [[1,1], [2,1]])
            sage: TestSuite(RC).run()
            sage: RC = RiggedConfigurations(['D', 4, 1], [[2,1], [1,1]])
            sage: TestSuite(RC).run() # long time
            sage: RC = RiggedConfigurations(['D', 4, 1], [[3,1]])
            sage: TestSuite(RC).run() # long time
            sage: RC = RiggedConfigurations(['D', 4, 1], [[4,2]])
            sage: TestSuite(RC).run() # long time
        """
        self._cartan_type = cartan_type
        self.dims = B
        cl = cartan_type.classical()
        self._rc_index = cl.index_set()
        self._rc_index_inverse = {i: ii for ii,i in enumerate(self._rc_index)}
        # We store the Cartan matrix for the vacancy number calculations for speed
        self._cartan_matrix = cl.cartan_matrix()
        Parent.__init__(self, category=KirillovReshetikhinCrystals().TensorProducts())

    # add options to class
    class options(GlobalOptions):
        r"""
        Set and display the options for rigged configurations.
        If no parameters are set, then the function returns a copy of
        the options dictionary.

        The ``options`` to partitions can be accessed as the method
        :obj:`RiggedConfigurations.options` of
        :class:`RiggedConfigurations`.

        @OPTIONS@

        EXAMPLES::

            sage: RC = RiggedConfigurations(['A',3,1], [[2,2],[1,1],[1,1]])
            sage: elt = RC(partition_list=[[3,1], [3], [1]])
            sage: elt
            <BLANKLINE>
            -3[ ][ ][ ]-3
            -1[ ]-1
            <BLANKLINE>
            1[ ][ ][ ]1
            <BLANKLINE>
            -1[ ]-1
            <BLANKLINE>
            sage: RiggedConfigurations.options(display='horizontal', convention='french')
            sage: elt
            -1[ ]-1         1[ ][ ][ ]1   -1[ ]-1
            -3[ ][ ][ ]-3

        Changing the ``convention`` for rigged configurations also changes the
        ``convention`` option for tableaux and vice versa::

            sage: T = Tableau([[1,2,3],[4,5]])
            sage: T.pp()
              4  5
              1  2  3
            sage: Tableaux.options.convention="english"
            sage: elt
            -3[ ][ ][ ]-3   1[ ][ ][ ]1   -1[ ]-1
            -1[ ]-1
            sage: T.pp()
              1  2  3
              4  5
            sage: RiggedConfigurations.options._reset()
        """
        NAME = 'RiggedConfigurations'
        module = 'sage.combinat.rigged_configurations.rigged_configurations'
        display = dict(default='vertical',
                     description='Specifies how rigged configurations should be printed',
                     values=dict(vertical='displayed vertically',
                                 horizontal='displayed horizontally'),
                     case_sensitive=False)
        element_ascii_art = dict(default=True,
                         description='display using the repr option ``element_ascii_art``',
                         checker=lambda x: isinstance(x, bool))
        half_width_boxes_type_B = dict(default=True,
                description='display the last rigged partition in affine type B as half width boxes',
                checker=lambda x: isinstance(x, bool))
        convention = dict(link_to=(tableau.Tableaux.options,'convention'))
        notation = dict(alt_name='convention')

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: RiggedConfigurations(['A', 3, 1], [[3, 2], [1, 2], [1, 1]])
            Rigged configurations of type ['A', 3, 1] and factor(s) ((3, 2), (1, 2), (1, 1))
        """
        return "Rigged configurations of type {} and factor(s) {}".format(self._cartan_type, self.dims)

    def _repr_option(self, key):
        """
        Metadata about the :meth:`_repr_` output.

        See :meth:`sage.structure.parent._repr_option` for details.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['A', 3, 1], [[2,1]])
            sage: RC._repr_option('element_ascii_art')
            True
        """
        if key == 'element_ascii_art':
            return self.options.element_ascii_art
        return super()._repr_option(key)

    def __iter__(self):
        """
        Iterate over ``self``.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['A', 3, 1], [[2,1], [1,1]])
            sage: L = [x for x in RC]
            sage: len(L)
            24
        """
        index_set = self._rc_index
        from sage.sets.recursively_enumerated_set import RecursivelyEnumeratedSet
        return RecursivelyEnumeratedSet(self.module_generators,
                    lambda x: [x.f(i) for i in index_set],
                    structure='graded').breadth_first_search_iterator()

    @lazy_attribute
    def module_generators(self):
        r"""
        Module generators for this set of rigged configurations.

        Iterate over the highest weight rigged configurations by moving
        through the
        :class:`~sage.combinat.rigged_configurations.kleber_tree.KleberTree`
        and then setting appropriate values of the partitions.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['D', 4, 1], [[2,1]])
            sage: for x in RC.module_generators: x
            <BLANKLINE>
            (/)
            <BLANKLINE>
            (/)
            <BLANKLINE>
            (/)
            <BLANKLINE>
            (/)
            <BLANKLINE>
            <BLANKLINE>
            0[ ]0
            <BLANKLINE>
            0[ ]0
            0[ ]0
            <BLANKLINE>
            0[ ]0
            <BLANKLINE>
            0[ ]0
            <BLANKLINE>

        TESTS:

        We check that this works with relabelled Cartan types (:issue:`16876`)::

            sage: ct = CartanType(['A',3,1]).relabel(lambda x: x+2)
            sage: RC = RiggedConfigurations(ct, [[4,1],[5,1]])
            sage: len(RC.module_generators)
            2
            sage: ct = CartanType(['A',3,1]).relabel(lambda x: (x+2) % 4)
            sage: RC = RiggedConfigurations(ct, [[0,1],[1,1]])
            sage: len(RC.module_generators)
            2
        """
        module_gens = []
        n = len(self._rc_index)

        for tree_node in self.kleber_tree():
            shapes = []
            cur = tree_node
            path_lambda = [cur.up_root.to_vector()] # Build the lambda values
            # Note that these are not same lambda as in the paper,
            #   but a less computational version.
            while cur.parent_node is not None:
                path_lambda.insert(0, (cur.parent_node.up_root - cur.up_root).to_vector())
                cur = cur.parent_node

            for a in range(n):
                shapes.append([])
                for i, cur_lambda in enumerate(path_lambda):
                    for j in range(cur_lambda[a]):
                        shapes[-1].insert(0, i)

            # Start with a base to calculate the vacancy numbers
            # Make a copy just to be safe
            base = self.element_class(self, partition_list=shapes[:])

            # Build out the blocks for the partition values
            vac_nums = []
            blocks = []
            L = []

            for partition in base:
                vac_nums.append(partition.vacancy_numbers)
                blocks.append([[]])

                # If the partition is empty, there's nothing to do
                if not partition._list:
                    L.append([[]])
                    continue

                # Setup the first block
                block_len = partition[0]
                for i, rowLen in enumerate(partition):
                    # If we've gone to a different sized block, then update the
                    #   values which change when moving to a new block size
                    if block_len != rowLen:
                        blocks[-1].append([])
                        block_len = rowLen

                    blocks[-1][-1].append(partition.vacancy_numbers[i])

                L2 = []
                for block in blocks[-1]:
                    L2.append(IterableFunctionCall(self._block_iterator, block))
                L.append(itertools.product(*L2))

            C = itertools.product(*L)
            for curBlocks in C:
                module_gens.append( self.element_class(self, KT_constructor=[shapes[:],
                                         self._blocks_to_values(curBlocks[:]),
                                         vac_nums[:]]) )

        return tuple(module_gens)

    def _block_iterator(self, container):
        r"""
        Iterate over all possible riggings for a particular block.

        Helper iterator which iterates over all possible partitions contained
        within the container.

        INPUT:

        - ``container`` -- list of widths of the rows of the container

        TESTS::

            sage: RC = RiggedConfigurations(['A', 4, 1], [[2, 2]])
            sage: for x in RC._block_iterator([]): x
            []
            sage: for x in RC._block_iterator([2,3]): x
            [0, 0]
            [1, 0]
            [1, 1]
            [2, 0]
            [2, 1]
            [2, 2]
        """
        if not container:
            yield []
            return

        pos = 0
        length = len(container)
        ret_part = [-1] * length
        while pos >= 0:
            ret_part[pos] += 1

            if ret_part[pos] > container[pos] or (pos != 0 and ret_part[pos] > ret_part[pos - 1]):
                ret_part[pos] = -1
                pos -= 1
            else:
                pos += 1

            if pos == length:
                yield ret_part[:]
                pos -= 1

    def _blocks_to_values(self, blocks):
        r"""
        Convert an array of blocks into a list of partition values.

        INPUT:

        - ``blocks`` -- the (2-dim) array blocks of the partition values

        TESTS::

            sage: RC = RiggedConfigurations(['A', 4, 1], [[2, 2]])
            sage: RC._blocks_to_values([[[2, 1]]])
            [[2, 1]]
        """
        values = []
        for part_block in blocks:
            if not part_block:
                values.append([])
            else:
                values.append(part_block[0][:]) # Need to make a copy
                for block in part_block[1:]:
                    values[-1].extend(block)
        return values

    def classically_highest_weight_vectors(self):
        """
        Return the classically highest weight elements of ``self``.

        TESTS::

            sage: RC = RiggedConfigurations(['A', 4, 1], [[2, 2]])
            sage: ascii_art(RC.classically_highest_weight_vectors())
            ( (/)  (/)  (/)  (/) )
        """
        return self.module_generators

    def _element_constructor_(self, *lst, **options):
        """
        Construct a ``RiggedConfigurationElement``.

        Typically the user should not call this method since it does not check
        if it is a valid configuration. Instead the user should use the
        iterator methods.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['A', 4, 1], [[2, 1]])
            sage: RC(partition_list=[[1], [1], [], []], rigging_list=[[-1], [0], [], []])
            <BLANKLINE>
            -1[ ]-1
            <BLANKLINE>
            0[ ]0
            <BLANKLINE>
            (/)
            <BLANKLINE>
            (/)
            <BLANKLINE>

        TESTS::

            sage: KT = crystals.TensorProductOfKirillovReshetikhinTableaux(['C',2,1], [[2,4],[1,2]])
            sage: t = KT(pathlist=[[2,1,2,1,-2,2,-1,-2],[2,-2]])
            sage: rc = t.to_rigged_configuration(); rc
            <BLANKLINE>
            -1[ ][ ][ ]-1
             0[ ][ ]0
            <BLANKLINE>
            -1[ ][ ]-1
            -1[ ]-1
            -1[ ]-1
            <BLANKLINE>
            sage: RC = RiggedConfigurations(['C',2,1], [[1,2],[2,4]])
            sage: RC(rc)
            <BLANKLINE>
            -1[ ][ ][ ]-1
             0[ ][ ]0
            <BLANKLINE>
            -1[ ][ ]-1
            -1[ ]-1
            -1[ ]-1
            <BLANKLINE>

        TESTS:

        Check that :issue:`17054` is fixed::

            sage: B = crystals.infinity.RiggedConfigurations(['A',2])
            sage: RC = RiggedConfigurations(['A',2,1], [[1,1]]*4 + [[2,1]]*4)
            sage: x = B.an_element().f_string([2,2,1,1,2,1,2,1])
            sage: ascii_art(x)
            -4[ ][ ][ ][ ]-4  -4[ ][ ][ ][ ]0
            sage: ascii_art(RC(x))
            0[ ][ ][ ][ ]-4  0[ ][ ][ ][ ]0
            sage: x == B.an_element().f_string([2,2,1,1,2,1,2,1])
            True
        """
        if not lst:
            return self.element_class(self, [], **options)

        from sage.combinat.rigged_configurations.tensor_product_kr_tableaux_element import TensorProductOfKirillovReshetikhinTableauxElement
        if isinstance(lst[0], TensorProductOfKirillovReshetikhinTableauxElement):
            if self != lst[0].parent().rigged_configurations():
                raise ValueError("incorrect bijection image")
            return lst[0].to_rigged_configuration()

        from sage.combinat.crystals.tensor_product import TensorProductOfRegularCrystalsElement
        if isinstance(lst[0], TensorProductOfRegularCrystalsElement):
            lst = lst[0]
        from sage.combinat.crystals.kirillov_reshetikhin import KirillovReshetikhinGenericCrystalElement
        if isinstance(lst[0], KirillovReshetikhinGenericCrystalElement):
            KRT = self.tensor_product_of_kirillov_reshetikhin_tableaux()
            krt_elt = KRT(*[x.to_kirillov_reshetikhin_tableau() for x in lst])
            return krt_elt.to_rigged_configuration()

        if isinstance(lst[0], (list, tuple)):
            lst = lst[0]

        if isinstance(lst[0], RiggedPartition):
            lst = [p._clone() for p in lst] # Make a deep copy
        elif isinstance(lst[0], RiggedConfigurationElement):
            lst = [p._clone() for p in lst[0]] # Make a deep copy

        return self.element_class(self, list(lst), **options)

    def _calc_vacancy_number(self, partitions, a, i, **options):
        r"""
        Calculate the vacancy number `p_i^{(a)}` in ``self``.

        This assumes that `\gamma_a = 1` for all `a` and `(\alpha_a \mid
        \alpha_b ) = A_{ab}`.

        INPUT:

        - ``partitions`` -- the list of rigged partitions we are using

        - ``a`` -- the rigged partition index

        - ``i`` -- the row length

        TESTS::

            sage: RC = RiggedConfigurations(['A', 4, 1], [[2, 1]])
            sage: elt = RC(partition_list=[[1], [1], [], []])
            sage: RC._calc_vacancy_number(elt.nu(), 1, 1)
            0
        """
        vac_num = 0
        if "B" in options:
            for tab in options["B"]:
                if len(tab) == self._rc_index[a]:
                    vac_num += min(i, len(tab[0]))
        elif "L" in options:
            L = options["L"]
            if a in L:
                for kvp in L[a].items():
                    vac_num += min(kvp[0], i) * kvp[1]
        elif "dims" in options:
            for dim in options["dims"]:
                if dim[0] == self._rc_index[a]:
                    vac_num += min(dim[1], i)
        else:
            for dim in self.dims:
                if dim[0] == self._rc_index[a]:
                    vac_num += min(dim[1], i)

        if i == float('inf'):
            vac_num -= sum(self._cartan_matrix[a,b] * sum(nu)
                           for b,nu in enumerate(partitions))
        else:
            vac_num -= sum(self._cartan_matrix[a,b] * nu.get_num_cells_to_column(i)
                           for b,nu in enumerate(partitions))

        return vac_num

    def kleber_tree(self):
        r"""
        Return the underlying Kleber tree used to generate all highest
        weight rigged configurations.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['A',3,1], [[1,1], [2,1]])
            sage: RC.kleber_tree()
            Kleber tree of Cartan type ['A', 3, 1] and B = ((1, 1), (2, 1))
        """
        return KleberTree(self._cartan_type, self.dims)

    @cached_method
    def tensor_product_of_kirillov_reshetikhin_tableaux(self):
        """
        Return the corresponding tensor product of Kirillov-Reshetikhin
        tableaux.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['A', 3, 1], [[3, 2], [1, 2]])
            sage: RC.tensor_product_of_kirillov_reshetikhin_tableaux()
            Tensor product of Kirillov-Reshetikhin tableaux of type ['A', 3, 1] and factor(s) ((3, 2), (1, 2))
        """
        from sage.combinat.rigged_configurations.tensor_product_kr_tableaux import TensorProductOfKirillovReshetikhinTableaux
        return TensorProductOfKirillovReshetikhinTableaux(self._cartan_type, self.dims)

    @cached_method
    def tensor_product_of_kirillov_reshetikhin_crystals(self):
        """
        Return the corresponding tensor product of Kirillov-Reshetikhin
        crystals.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['A', 3, 1], [[3,1],[2,2]])
            sage: RC.tensor_product_of_kirillov_reshetikhin_crystals()
            Full tensor product of the crystals
            [Kirillov-Reshetikhin crystal of type ['A', 3, 1] with (r,s)=(3,1),
             Kirillov-Reshetikhin crystal of type ['A', 3, 1] with (r,s)=(2,2)]
        """
        return self.tensor_product_of_kirillov_reshetikhin_tableaux().tensor_product_of_kirillov_reshetikhin_crystals()

    def fermionic_formula(self, q=None, only_highest_weight=False, weight=None):
        r"""
        Return the fermionic formula associated to ``self``.

        Given a set of rigged configurations `RC(\lambda, L)`, the fermionic
        formula is defined as:

        .. MATH::

            M(\lambda, L; q) = \sum_{(\nu,J)} q^{cc(\nu, J)}

        where we sum over all (classically highest weight) rigged
        configurations of weight `\lambda` where `cc` is the
        :meth:`cocharge statistic
        <sage.combinat.rigged_configurations.rigged_configuration_element.RiggedConfigurationElement.cc>`.
        This is known to reduce to

        .. MATH::

            M(\lambda, L; q) = \sum_{\nu} q^{cc(\nu)} \prod_{(a,i) \in
            I \times \ZZ} \begin{bmatrix} p_i^{(a)} + m_i^{(a)} \\ m_i^{(a)}
            \end{bmatrix}_q.

        The generating function of `M(\lambda, L; q)` in the weight algebra
        subsumes all fermionic formulas:

        .. MATH::

            M(L; q) = \sum_{\lambda \in P} M(\lambda, L; q) \lambda.

        This is conjecturally equal to the
        :meth:`one dimensional configuration sum
        <sage.combinat.crystals.tensor_product.CrystalOfWords.one_dimensional_configuration_sum>`
        of the corresponding tensor product of Kirillov-Reshetikhin crystals, see [HKOTT2002]_.
        This has been proven in general for type `A_n^{(1)}` [BijectionLRT]_,
        single factors `B^{r,s}` in type `D_n^{(1)}` [OSS2011]_ with the result
        from [Sakamoto13]_, as well as for a tensor product of single columns
        [OSS2003]_, [BijectionDn]_ or a tensor product of single rows [OSS03]_
        for all non-exceptional types.

        INPUT:

        - ``q`` -- the variable `q`
        - ``only_highest_weight`` -- use only the classically highest weight
          rigged configurations
        - ``weight`` -- return the fermionic formula `M(\lambda, L; q)` where
          `\lambda` is the classical weight ``weight``

        REFERENCES:

        .. [OSS2003] Masato Okado, Anne Schilling, and Mark Shimozono.
           Virtual crystals and fermionic formulas of type `D_{n+1}^{(2)}`,
           `A_{2n}^{(2)}`, and `C_n^{(1)}`. Representation Theory. **7** (2003)
           :arxiv:`math.QA/0105017`.

        .. [Sakamoto13] Reiho Sakamoto.
           Rigged configurations and Kashiwara operators.
           (2013) :arxiv:`1302.4562v1`.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['A', 2, 1], [[1,1], [1,1]])
            sage: RC.fermionic_formula()
            B[-2*Lambda[1] + 2*Lambda[2]] + (q+1)*B[-Lambda[1]]
             + (q+1)*B[Lambda[1] - Lambda[2]] + B[2*Lambda[1]]
             + B[-2*Lambda[2]] + (q+1)*B[Lambda[2]]
            sage: t = QQ['t'].gen(0)
            sage: RC.fermionic_formula(t)
            B[-2*Lambda[1] + 2*Lambda[2]] + (t+1)*B[-Lambda[1]]
             + (t+1)*B[Lambda[1] - Lambda[2]] + B[2*Lambda[1]]
             + B[-2*Lambda[2]] + (t+1)*B[Lambda[2]]
            sage: La = RC.weight_lattice_realization().classical().fundamental_weights()
            sage: RC.fermionic_formula(weight=La[2])
            q + 1
            sage: RC.fermionic_formula(only_highest_weight=True, weight=La[2])
            q

        Only using the highest weight elements on other types::

            sage: RC = RiggedConfigurations(['A', 3, 1], [[3,1], [2,2]])
            sage: RC.fermionic_formula(only_highest_weight=True)
            q*B[Lambda[1] + Lambda[2]] + B[2*Lambda[2] + Lambda[3]]
            sage: RC = RiggedConfigurations(['D', 4, 1], [[3,1], [4,1], [2,1]])
            sage: RC.fermionic_formula(only_highest_weight=True)
            (q^4+q^3+q^2)*B[Lambda[1]] + (q^2+q)*B[Lambda[1] + Lambda[2]]
             + q*B[Lambda[1] + 2*Lambda[3]] + q*B[Lambda[1] + 2*Lambda[4]]
             + B[Lambda[2] + Lambda[3] + Lambda[4]] + (q^3+2*q^2+q)*B[Lambda[3] + Lambda[4]]
            sage: RC = RiggedConfigurations(['E', 6, 1], [[2,2]])
            sage: RC.fermionic_formula(only_highest_weight=True)
            q^2*B[0] + q*B[Lambda[2]] + B[2*Lambda[2]]
            sage: RC = RiggedConfigurations(['B', 3, 1], [[3,1], [2,2]])
            sage: RC.fermionic_formula(only_highest_weight=True) # long time
            q*B[Lambda[1] + Lambda[2] + Lambda[3]] + q^2*B[Lambda[1]
             + Lambda[3]] + (q^2+q)*B[Lambda[2] + Lambda[3]] + B[2*Lambda[2]
             + Lambda[3]] + (q^3+q^2)*B[Lambda[3]]
            sage: RC = RiggedConfigurations(['C', 3, 1], [[3,1], [2,2]])
            sage: RC.fermionic_formula(only_highest_weight=True) # long time
            (q^3+q^2)*B[Lambda[1] + Lambda[2]] + q*B[Lambda[1] + 2*Lambda[2]]
             + (q^2+q)*B[2*Lambda[1] + Lambda[3]] + B[2*Lambda[2] + Lambda[3]]
             + (q^4+q^3+q^2)*B[Lambda[3]]
            sage: RC = RiggedConfigurations(['D', 4, 2], [[3,1], [2,2]])
            sage: RC.fermionic_formula(only_highest_weight=True) # long time
            (q^2+q)*B[Lambda[1] + Lambda[2] + Lambda[3]] + (q^5+2*q^4+q^3)*B[Lambda[1]
             + Lambda[3]] + (q^3+q^2)*B[2*Lambda[1] + Lambda[3]] + (q^4+q^3+q^2)*B[Lambda[2]
             + Lambda[3]] + B[2*Lambda[2] + Lambda[3]] + (q^6+q^5+q^4)*B[Lambda[3]]
            sage: RC = RiggedConfigurations(CartanType(['A',4,2]).dual(), [[1,1],[2,2]])
            sage: RC.fermionic_formula(only_highest_weight=True)
            (q^3+q^2)*B[Lambda[1]] + (q^2+q)*B[Lambda[1] + 2*Lambda[2]]
             + B[Lambda[1] + 4*Lambda[2]] + q*B[3*Lambda[1]] + q*B[4*Lambda[2]]

        TESTS::

            sage: RC = RiggedConfigurations(['A', 2, 1], [[1,1], [1,1]])
            sage: KR = RC.tensor_product_of_kirillov_reshetikhin_crystals()
            sage: RC.fermionic_formula() == KR.one_dimensional_configuration_sum()
            True
            sage: KT = RC.tensor_product_of_kirillov_reshetikhin_tableaux()
            sage: RC.fermionic_formula() == KT.one_dimensional_configuration_sum()
            True
            sage: RC = RiggedConfigurations(['C', 2, 1], [[2,1], [2,1]])
            sage: KR = RC.tensor_product_of_kirillov_reshetikhin_crystals()
            sage: RC.fermionic_formula() == KR.one_dimensional_configuration_sum() # long time
            True
            sage: t = QQ['t'].gen(0)
            sage: RC = RiggedConfigurations(['D', 4, 1], [[1,1], [2,1]])
            sage: KR = RC.tensor_product_of_kirillov_reshetikhin_crystals()
            sage: RC.fermionic_formula(t) == KR.one_dimensional_configuration_sum(t) # long time
            True
        """
        if q is None:
            from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
            q = PolynomialRing(QQ, 'q').gen(0)

        if only_highest_weight:
            L = self.module_generators
        else:
            L = self

        P = q.parent()
        WLR = self.weight_lattice_realization().classical()

        if weight is not None:
            weight = WLR(weight)
            return P.sum(q**x.cc() for x in L if WLR(x.weight()) == weight)

        B = WLR.algebra(P)
        return B.sum(q**x.cc() * B(WLR(x.weight())) for x in L)

    def _test_bijection(self, **options):
        r"""
        Test function to make sure that the bijection between rigged
        configurations and Kirillov-Reshetikhin tableaux is correct.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['A', 3, 1], [[2,1],[1,1]])
            sage: RC._test_bijection()
        """
        tester = self._tester(**options)
        rejects = []
        for x in self:
            y = x.to_tensor_product_of_kirillov_reshetikhin_tableaux()
            z = y.to_rigged_configuration()
            if z != x:
                rejects.append((x, z))

        tester.assertEqual(len(rejects), 0, "Bijection is not correct: {}".format(rejects))
        if rejects:
            return rejects

    def tensor(self, *crystals, **options):
        """
        Return the tensor product of ``self`` with ``crystals``.

        If ``crystals`` is a list of rigged configurations of the same
        Cartan type, then this returns a new :class:`RiggedConfigurations`.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['A', 3, 1], [[2,1],[1,3]])
            sage: RC2 = RiggedConfigurations(['A', 3, 1], [[1,1], [3,3]])
            sage: RC.tensor(RC2, RC2)
            Rigged configurations of type ['A', 3, 1]
             and factor(s) ((2, 1), (1, 3), (1, 1), (3, 3), (1, 1), (3, 3))

            sage: K = crystals.KirillovReshetikhin(['A', 3, 1], 2, 2, model='KR')
            sage: RC.tensor(K)
            Full tensor product of the crystals
             [Rigged configurations of type ['A', 3, 1] and factor(s) ((2, 1), (1, 3)),
              Kirillov-Reshetikhin tableaux of type ['A', 3, 1] and shape (2, 2)]
        """
        ct = self._cartan_type
        if all(isinstance(B, RiggedConfigurations) and B.cartan_type() == ct for B in crystals):
            dims = self.dims
            for B in crystals:
                dims += B.dims
            return RiggedConfigurations(ct, dims)
        return super().tensor(*crystals, **options)

    Element = KRRCSimplyLacedElement


class RCNonSimplyLaced(RiggedConfigurations):
    r"""
    Rigged configurations in non-simply-laced types.

    These are rigged configurations which lift to virtual rigged configurations
    in a simply-laced type.

    For more on rigged configurations, see :class:`RiggedConfigurations`.
    """
    @staticmethod
    def __classcall_private__(cls, cartan_type, B):
        r"""
        Normalize the input arguments to ensure unique representation.

        EXAMPLES::

            sage: RC1 = RiggedConfigurations(CartanType(['A',4,2]), [[2,2]])
            sage: RC2 = RiggedConfigurations(['A',4,2], [(2,2)])
            sage: RC3 = RiggedConfigurations(['BC',2,2], ((2,2),))
            sage: RC2 is RC1, RC3 is RC1
            (True, True)
        """
        cartan_type = CartanType(cartan_type)

        # Standardize B input into a tuple of tuples
        B = tuple(map(tuple, B))
        return super().__classcall__(cls, cartan_type, B)

    def __init__(self, cartan_type, dims):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['C',2,1], [[1,1]])
            sage: TestSuite(RC).run()
            sage: RC = RiggedConfigurations(['C',2,1], [[1,2],[2,1]])
            sage: TestSuite(RC).run() # long time
            sage: RC = RiggedConfigurations(['B',3,1], [[3,1],[1,1]])
            sage: TestSuite(RC).run() # long time
            sage: RC = RiggedConfigurations(['D',4,2], [[2,1]])
            sage: TestSuite(RC).run() # long time
            sage: RC = RiggedConfigurations(['A',5,2], [[2,1]])
            sage: TestSuite(RC).run() # long time
        """
        self._folded_ct = cartan_type.as_folding()
        RiggedConfigurations.__init__(self, cartan_type, dims)

    def _calc_vacancy_number(self, partitions, a, i, **options):
        r"""
        Calculate the vacancy number `p_i^{(a)}` in ``self``.

        INPUT:

        - ``partitions`` -- the list of rigged partitions we are using

        - ``a`` -- the rigged partition index

        - ``i`` -- the row length

        TESTS::

            sage: RC = RiggedConfigurations(['C', 4, 1], [[2, 1]])
            sage: elt = RC(partition_list=[[1], [2], [2], [1]])
            sage: RC._calc_vacancy_number(elt.nu(), 1, 2)
            0
        """
        vac_num = 0
        if "B" in options:
            for tab in options["B"]:
                if len(tab) == self._rc_index[a]:
                    vac_num += min(i, len(tab[0]))
        elif "L" in options:
            L = options["L"]
            if a in L:
                for kvp in L[a].items():
                    vac_num += min(kvp[0], i) * kvp[1]
        elif "dims" in options:
            for dim in options["dims"]:
                if dim[0] == self._rc_index[a]:
                    vac_num += min(dim[1], i)
        else:
            for dim in self.dims:
                if dim[0] == self._rc_index[a]:
                    vac_num += min(dim[1], i)

        if i == float('inf'):
            vac_num -= sum(self._cartan_matrix[a,b] * sum(nu)
                           for b,nu in enumerate(partitions))
        else:
            gamma = self._folded_ct.scaling_factors()
            vac_num -= sum(self._cartan_matrix[a,b]
                           * nu.get_num_cells_to_column(gamma[a+1]*i, gamma[b+1])
                           // gamma[b+1]
                           for b,nu in enumerate(partitions))

        return vac_num

    @lazy_attribute
    def module_generators(self):
        r"""
        Module generators for this set of rigged configurations.

        Iterate over the highest weight rigged configurations by moving
        through the
        :class:`~sage.combinat.rigged_configurations.kleber_tree.KleberTree`
        and then setting appropriate values of the partitions.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['C', 3, 1], [[1,2]])
            sage: for x in RC.module_generators: x
            <BLANKLINE>
            (/)
            <BLANKLINE>
            (/)
            <BLANKLINE>
            (/)
            <BLANKLINE>
            <BLANKLINE>
            0[ ][ ]0
            <BLANKLINE>
            0[ ][ ]0
            <BLANKLINE>
            0[ ]0
            <BLANKLINE>

            sage: RC = RiggedConfigurations(['D',4,3], [[1,1]])
            sage: RC.module_generators
            (
            <BLANKLINE>
                 0[ ]0
            (/)  0[ ]0
            <BLANKLINE>
            (/)  0[ ]0
               ,
            )
        """
        module_gens = []
        vec_len = len(self.kleber_tree().root.up_root.to_vector())

        for tree_node in self.kleber_tree():
            shapes = []
            cur = tree_node
            path_lambda = [cur.up_root.to_vector()] # Build the lambda values
            # Note that these are not same lambda as in the paper,
            #   but a less computational version.
            while cur.parent_node is not None:
                path_lambda.insert(0, (cur.parent_node.up_root - cur.up_root).to_vector())
                cur = cur.parent_node

            for a in range(vec_len):
                shapes.append([])
                for i, cur_lambda in enumerate(path_lambda):
                    for j in range(cur_lambda[a]):
                        shapes[-1].insert(0, i)

            # Convert from the virtual rigged configuration
            # As a special case, we do not need to do anything for type `A_{2n}^{(2)}`
            sigma = self._folded_ct.folding_orbit()
            vindex = self.virtual._rc_index
            shapes = [shapes[vindex.index(sigma[a][0])] for a in self._rc_index]
            if self._cartan_type.type() != 'BC':
                gamma = self._folded_ct.scaling_factors()
                for a,shape in enumerate(shapes):
                    for i in range(len(shape)):
                        shape[i] = shape[i] // gamma[self._rc_index[a]]

            # Start with a base to calculate the vacancy numbers
            # Make a copy just to be safe
            base = self.element_class(self, partition_list=shapes[:])

            # Build out the blocks for the partition values
            vac_nums = []
            blocks = []
            L = []

            for partition in base:
                vac_nums.append(partition.vacancy_numbers)
                blocks.append([[]])

                # If the partition is empty, there's nothing to do
                if not partition._list:
                    L.append([[]])
                    continue

                # Setup the first block
                block_len = partition[0]
                for i, rowLen in enumerate(partition):
                    # If we've gone to a different sized block, then update the
                    #   values which change when moving to a new block size
                    if block_len != rowLen:
                        blocks[-1].append([])
                        block_len = rowLen

                    blocks[-1][-1].append(partition.vacancy_numbers[i])

                L2 = []
                for block in blocks[-1]:
                    L2.append(IterableFunctionCall(self._block_iterator, block))
                L.append(itertools.product(*L2))

            C = itertools.product(*L)
            for cur_blocks in C:
                module_gens.append( self.element_class(self, KT_constructor=[shapes[:],
                                         self._blocks_to_values(cur_blocks[:]), vac_nums[:]]) )

        return tuple(module_gens)

    def kleber_tree(self):
        r"""
        Return the underlying (virtual) Kleber tree used to generate all
        highest weight rigged configurations.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['C',3,1], [[1,1], [2,1]])
            sage: RC.kleber_tree()
            Virtual Kleber tree of Cartan type ['C', 3, 1] and B = ((1, 1), (2, 1))
        """
        return VirtualKleberTree(self._cartan_type, self.dims)

    @lazy_attribute
    def virtual(self):
        """
        Return the corresponding virtual crystal.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['C',2,1], [[1,2],[1,1],[2,1]])
            sage: RC
            Rigged configurations of type ['C', 2, 1] and factor(s) ((1, 2), (1, 1), (2, 1))
            sage: RC.virtual
            Rigged configurations of type ['A', 3, 1] and factor(s) ((1, 2), (3, 2), (1, 1), (3, 1), (2, 2))
        """
        gamma = self._folded_ct.scaling_factors()
        sigma = self._folded_ct.folding_orbit()
        virtual_dims = []
        for r,s in self.dims:
            for a in sigma[r]:
                virtual_dims.append([a, s*gamma[r]])
        return RiggedConfigurations(self._folded_ct._folding, virtual_dims)

    def to_virtual(self, rc):
        """
        Convert ``rc`` into a rigged configuration in the virtual crystal.

        INPUT:

        - ``rc`` -- a rigged configuration element

        EXAMPLES::

            sage: RC = RiggedConfigurations(['C',2,1], [[1,2],[1,1],[2,1]])
            sage: elt = RC(partition_list=[[3],[2]]); elt
            <BLANKLINE>
            0[ ][ ][ ]0
            <BLANKLINE>
            0[ ][ ]0
            sage: velt = RC.to_virtual(elt); velt
            <BLANKLINE>
            0[ ][ ][ ]0
            <BLANKLINE>
            0[ ][ ][ ][ ]0
            <BLANKLINE>
            0[ ][ ][ ]0
            sage: velt.parent()
            Rigged configurations of type ['A', 3, 1] and factor(s) ((1, 2), (3, 2), (1, 1), (3, 1), (2, 2))
        """
        gamma = self._folded_ct.scaling_factors()
        sigma = self._folded_ct.folding_orbit()
        n = len(self.virtual._rc_index)
        # +/- 1 for indexing
        partitions = [None] * n
        for a,rp in enumerate(rc):
            g = gamma[a+1]
            for i in sigma[a+1]:
                partitions[i-1] = RiggedPartition([row_len*g for row_len in rp._list],
                                                  [rig_val*g for rig_val in rp.rigging],
                                                  [vac_num*g for vac_num in rp.vacancy_numbers])
        return self.virtual.element_class(self.virtual, partitions, use_vacancy_numbers=True)

    def from_virtual(self, vrc):
        """
        Convert ``vrc`` in the virtual crystal into a rigged configuration of
        the original Cartan type.

        INPUT:

        - ``vrc`` -- a virtual rigged configuration

        EXAMPLES::

            sage: RC = RiggedConfigurations(['C',2,1], [[1,2],[1,1],[2,1]])
            sage: elt = RC(partition_list=[[3],[2]])
            sage: vrc_elt = RC.to_virtual(elt)
            sage: ret = RC.from_virtual(vrc_elt); ret
            <BLANKLINE>
            0[ ][ ][ ]0
            <BLANKLINE>
            0[ ][ ]0
            sage: ret == elt
            True
        """
        gamma = self._folded_ct.scaling_factors()
        sigma = self._folded_ct.folding_orbit()
        n = len(self._rc_index)
        partitions = [None] * n
        # +/- 1 for indexing
        for a in range(n):
            rp = vrc[sigma[a+1][0] - 1]
            g = gamma[a+1]
            partitions[a] = RiggedPartition([row_len//g for row_len in rp._list],
                                            [rig_val//g for rig_val in rp.rigging],
                                            [vac_val//g for vac_val in rp.vacancy_numbers])
        return self.element_class(self, partitions, use_vacancy_numbers=True)

    def _test_virtual_vacancy_numbers(self, **options):
        """
        Test to make sure that the vacancy numbers obtained from the virtual
        rigged configuration agree with the explicit computation of the
        vacancy numbers done here.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['B', 3, 1], [[2,1]])
            sage: RC._test_virtual_vacancy_numbers()
        """
        tester = self._tester(**options)
        for x in self:
            parts_list = [p._list[:] for p in x]
            elt = self.element_class(self, partition_list=parts_list)
            for i, p in enumerate(elt):
                for j, vac_num in enumerate(p.vacancy_numbers):
                    tester.assertEqual(vac_num, x[i].vacancy_numbers[j],
                      "Incorrect vacancy number: {}\nComputed: {}\nFor: {}".format(
                       x[i].vacancy_numbers[j], vac_num, x))

    Element = KRRCNonSimplyLacedElement


class RCTypeA2Even(RCNonSimplyLaced):
    """
    Rigged configurations for type `A_{2n}^{(2)}`.

    For more on rigged configurations, see :class:`RiggedConfigurations`.

    EXAMPLES::

        sage: RC = RiggedConfigurations(['A',4,2], [[2,1], [1,2]])
        sage: RC.cardinality()
        150
        sage: RC = RiggedConfigurations(['A',2,2], [[1,1]])
        sage: RC.cardinality()
        3
        sage: RC = RiggedConfigurations(['A',2,2], [[1,2],[1,1]])
        sage: TestSuite(RC).run() # long time
        sage: RC = RiggedConfigurations(['A',4,2], [[2,1]])
        sage: TestSuite(RC).run() # long time
    """

    def cardinality(self):
        """
        Return the cardinality of ``self``.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['A',4,2], [[1,1], [2,2]])
            sage: RC.cardinality()
            250
        """
        return self.tensor_product_of_kirillov_reshetikhin_tableaux().cardinality()

    @lazy_attribute
    def virtual(self):
        """
        Return the corresponding virtual crystal.

        EXAMPLES::

            sage: RC = RiggedConfigurations(['A',4,2], [[1,2],[1,1],[2,1]])
            sage: RC
            Rigged configurations of type ['BC', 2, 2] and factor(s) ((1, 2), (1, 1), (2, 1))
            sage: RC.virtual
            Rigged configurations of type ['A', 3, 1] and factor(s) ((1, 2), (3, 2), (1, 1), (3, 1), (2, 1), (2, 1))
        """
        sigma = self._folded_ct.folding_orbit()
        n = len(sigma) - 1
        virtual_dims = []
        for r,s in self.dims:
            if r == n:
                virtual_dims.extend([[n, s], [n, s]])
            else:
                for a in sigma[r]:
                    virtual_dims.append([a, s])
        return RiggedConfigurations(self._folded_ct._folding, virtual_dims)

    def _calc_vacancy_number(self, partitions, a, i, **options):
        r"""
        Calculate the vacancy number `p_i^{(a)}` in ``self``.

        This is a special implementation for type `A_{2n}^{(2)}`.

        INPUT:

        - ``partitions`` -- the list of rigged partitions we are using

        - ``a`` -- the rigged partition index

        - ``i`` -- the row length

        TESTS::

            sage: RC = RiggedConfigurations(['A', 4, 2], [[2, 1]])
            sage: elt = RC(partition_list=[[1], [2]])
            sage: RC._calc_vacancy_number(elt.nu(), 1, 2)
            0
        """
        vac_num = 0
        if "B" in options:
            for tab in options["B"]:
                if len(tab) == self._rc_index[a]:
                    vac_num += min(i, len(tab[0]))
        elif "L" in options:
            L = options["L"]
            if a in L:
                for kvp in L[a].items():
                    vac_num += min(kvp[0], i) * kvp[1]
        elif "dims" in options:
            for dim in options["dims"]:
                if dim[0] == self._rc_index[a]:
                    vac_num += min(dim[1], i)
        else:
            for dim in self.dims:
                if dim[0] == self._rc_index[a]:
                    vac_num += min(dim[1], i)

        gamma = self._folded_ct.scaling_factors()
        if i == float('inf'):
            vac_num -= sum(self._cartan_matrix[a,b] * sum(nu) // gamma[b+1]
                           for b, nu in enumerate(partitions))
        else:
            vac_num -= sum(self._cartan_matrix[a,b] * nu.get_num_cells_to_column(i) // gamma[b+1]
                           for b, nu in enumerate(partitions))

        return vac_num

    def to_virtual(self, rc):
        """
        Convert ``rc`` into a rigged configuration in the virtual crystal.

        INPUT:

        - ``rc`` -- a rigged configuration element

        EXAMPLES::

            sage: RC = RiggedConfigurations(['A',4,2], [[2,2]])
            sage: elt = RC(partition_list=[[1],[1]]); elt
            <BLANKLINE>
            -1[ ]-1
            <BLANKLINE>
            1[ ]1
            <BLANKLINE>
            sage: velt = RC.to_virtual(elt); velt
            <BLANKLINE>
            -1[ ]-1
            <BLANKLINE>
            2[ ]2
            <BLANKLINE>
            -1[ ]-1
            <BLANKLINE>
            sage: velt.parent()
            Rigged configurations of type ['A', 3, 1] and factor(s) ((2, 2), (2, 2))
        """
        gamma = self._folded_ct.scaling_factors()
        sigma = self._folded_ct.folding_orbit()
        n = len(self.virtual._rc_index)
        partitions = [None] * n
        for a,rp in enumerate(rc):
            g = gamma[a+1]
            for i in sigma[a+1]:
                partitions[i-1] = RiggedPartition(list(rp._list),
                                                  [rig_val*g for rig_val in rp.rigging],
                                                  [vac_num*g for vac_num in rp.vacancy_numbers])
        return self.virtual.element_class(self.virtual, partitions, use_vacancy_numbers=True)

    def from_virtual(self, vrc):
        """
        Convert ``vrc`` in the virtual crystal into a rigged configuration of
        the original Cartan type.

        INPUT:

        - ``vrc`` -- a virtual rigged configuration element

        EXAMPLES::

            sage: RC = RiggedConfigurations(['A',4,2], [[2,2]])
            sage: elt = RC(partition_list=[[1],[1]])
            sage: velt = RC.to_virtual(elt)
            sage: ret = RC.from_virtual(velt); ret
            <BLANKLINE>
            -1[ ]-1
            <BLANKLINE>
            1[ ]1
            <BLANKLINE>
            sage: ret == elt
            True
        """
        gamma = self._folded_ct.scaling_factors()
        sigma = self._folded_ct.folding_orbit()
        n = len(self._rc_index)
        partitions = [None] * n
        # +/- 1 for indexing
        for a in range(n):
            rp = vrc[sigma[a+1][0] - 1]
            g = gamma[a+1]
            partitions[a] = RiggedPartition(list(rp._list),
                                            [rig_val//g for rig_val in rp.rigging],
                                            [vac_val//g for vac_val in rp.vacancy_numbers])
        return self.element_class(self, partitions, use_vacancy_numbers=True)


class RCTypeA2Dual(RCTypeA2Even):
    r"""
    Rigged configurations of type `A_{2n}^{(2)\dagger}`.

    For more on rigged configurations, see :class:`RiggedConfigurations`.

    EXAMPLES::

        sage: RC = RiggedConfigurations(CartanType(['A',4,2]).dual(), [[1,2],[1,1],[2,1]])
        sage: RC
        Rigged configurations of type ['BC', 2, 2]^* and factor(s) ((1, 2), (1, 1), (2, 1))
        sage: RC.cardinality()
        750
        sage: RC.virtual
        Rigged configurations of type ['A', 3, 1] and factor(s) ((1, 2), (3, 2), (1, 1), (3, 1), (2, 1), (2, 1))
        sage: RC = RiggedConfigurations(CartanType(['A',2,2]).dual(), [[1,1]])
        sage: RC.cardinality()
        3
        sage: RC = RiggedConfigurations(CartanType(['A',2,2]).dual(), [[1,2],[1,1]])
        sage: TestSuite(RC).run() # long time
        sage: RC = RiggedConfigurations(CartanType(['A',4,2]).dual(), [[2,1]])
        sage: TestSuite(RC).run() # long time
    """

    def _calc_vacancy_number(self, partitions, a, i, **options):
        r"""
        Calculate the vacancy number `p_i^{(a)}` in ``self``. A special case
        is needed for the `n`-th partition for type `A_{2n}^{(2)\dagger}`.

        INPUT:

        - ``partitions`` -- the list of rigged partitions we are using

        - ``a`` -- the rigged partition index

        - ``i`` -- the row length

        TESTS::

            sage: RC = RiggedConfigurations(CartanType(['A', 6, 2]).dual(), [[2,1]])
            sage: elt = RC(partition_list=[[1], [2], [2]])
            sage: RC._calc_vacancy_number(elt.nu(), 0, 1)
            -1
        """
        if a != len(self._rc_index) - 1:
            return RCTypeA2Even._calc_vacancy_number(self, partitions, a, i, **options)

        vac_num = 0
        if "B" in options:
            for tab in options["B"]:
                if len(tab) == self._rc_index[a]:
                    vac_num += min(i, len(tab[0]))
        elif "L" in options:
            L = options["L"]
            if a in L:
                for kvp in L[a].items():
                    vac_num += min(kvp[0], i) * kvp[1]
        elif "dims" in options:
            for dim in options["dims"]:
                if dim[0] == self._rc_index[a]:
                    vac_num += min(dim[1], i)
        else:
            for dim in self.dims:
                if dim[0] == self._rc_index[a]:
                    vac_num += min(dim[1], i)

        if i == float('inf'):
            vac_num -= sum(self._cartan_matrix[a,b] * sum(nu) / 2
                           for b,nu in enumerate(partitions))
        else:
            vac_num -= sum(self._cartan_matrix[a,b] * nu.get_num_cells_to_column(i) / 2
                           for b,nu in enumerate(partitions))

        return vac_num

    @lazy_attribute
    def module_generators(self):
        r"""
        Module generators for rigged configurations of type
        `A_{2n}^{(2)\dagger}`.

        Iterate over the highest weight rigged configurations by moving
        through the
        :class:`~sage.combinat.rigged_configurations.kleber_tree.KleberTree`
        and then setting appropriate values of the partitions. This also
        skips rigged configurations where `P_i^{(n)} < 1` when `i` is odd.

        EXAMPLES::

            sage: RC = RiggedConfigurations(CartanType(['A', 4, 2]).dual(), [[1,1]])
            sage: for x in RC.module_generators: x
            <BLANKLINE>
            (/)
            <BLANKLINE>
            (/)
            <BLANKLINE>
        """
        module_gens = []
        # This is for the non-simply-laced types
        vec_len = len(self.kleber_tree().root.up_root.to_vector())

        for tree_node in self.kleber_tree():
            shapes = []
            cur = tree_node
            path_lambda = [cur.up_root.to_vector()] # Build the lambda values
            # Note that these are not same lambda as in the paper,
            #   but a less computational version.
            while cur.parent_node is not None:
                path_lambda.insert(0, (cur.parent_node.up_root - cur.up_root).to_vector())
                cur = cur.parent_node

            for a in range(vec_len):
                shapes.append([])
                for i, cur_lambda in enumerate(path_lambda):
                    for j in range(cur_lambda[a]):
                        shapes[-1].insert(0, i)

            # We are not simply-laced, so convert from the virtual rigged configuration
            sigma = self._folded_ct.folding_orbit()
            vindex = self.virtual._rc_index
            shapes = [shapes[vindex.index(sigma[a][0])] for a in self._rc_index]
            # Nothing more to do since gamma[i] == 1 for all i >= 1

            # Start with a base to calculate the vacancy numbers
            # Make a copy just to be safe
            base = self.element_class(self, partition_list=shapes[:])

            # Check the special condition of odd rows in the n-th partition
            invalid_RC = False
            for i in range(len(base[-1]._list)):
                if base[-1]._list[i] % 2 == 1 and base[-1].vacancy_numbers[i] < 1:
                    invalid_RC = True
                    break
            # If it is invalid, skip it
            if invalid_RC:
                continue

            # Build out the blocks for the partition values
            vac_nums = []
            blocks = []
            L = []

            for partition in base[:-1]:
                vac_nums.append(partition.vacancy_numbers)
                blocks.append([[]])

                # If the partition is empty, there's nothing to do
                if len(partition) <= 0:
                    L.append([[]])
                    continue

                # Setup the first block
                block_len = partition[0]
                for i, row_len in enumerate(partition):
                    # If we've gone to a different sized block, then update the
                    #   values which change when moving to a new block size
                    if block_len != row_len:
                        blocks[-1].append([])
                        block_len = row_len

                    blocks[-1][-1].append(partition.vacancy_numbers[i])

                L2 = []
                for block in blocks[-1]:
                    L2.append(IterableFunctionCall(self._block_iterator, block))
                L.append(itertools.product(*L2))

            # Special case for the final tableau
            partition = base[-1]
            vac_nums.append(partition.vacancy_numbers)

            # If the partition is empty, there's nothing to do
            if len(partition) <= 0:
                L.append([[]])
            else:
                # Setup the first block
                block_len = partition[0]
                blocks = [[]]
                odd_block = []
                for i, row_len in enumerate(partition):
                    # If we've gone to a different sized block, then update the
                    #   values which change when moving to a new block size
                    if block_len != row_len:
                        blocks.append([])
                        odd_block.append(block_len % 2 == 1)
                        block_len = row_len

                    blocks[-1].append(partition.vacancy_numbers[i])
                odd_block.append(block_len % 2 == 1)

                L2 = []
                for i, block in enumerate(blocks):
                    if odd_block[i]:
                        L2.append(IterableFunctionCall(self._block_iterator_n_odd, block))
                    else:
                        L2.append(IterableFunctionCall(self._block_iterator, block))
                L.append(itertools.product(*L2))

            C = itertools.product(*L)
            for curBlocks in C:
                module_gens.append( self.element_class(self, KT_constructor=[shapes[:],
                                         self._blocks_to_values(curBlocks[:]), vac_nums[:]]) )

        return tuple(module_gens)

    def _block_iterator_n_odd(self, container):
        r"""
        Iterate over all possible riggings for a block of odd length in the
        `n`-th rigged partition for type `A_{2n}^{(2)\dagger}`.

        Helper iterator which iterates over all possible partitions of
        `\frac{2k+1}{2}` sizes contained within the container.

        INPUT:

        - ``container`` -- list the widths of the rows of the container

        TESTS::

            sage: RC = RiggedConfigurations(CartanType(['A', 4, 2]).dual(), [[2, 2]])
            sage: for x in RC._block_iterator_n_odd([]): x
            []
            sage: for x in RC._block_iterator_n_odd([2,2]): x
            [1/2, 1/2]
            [3/2, 1/2]
            [3/2, 3/2]
        """
        if len(container) == 0:
            yield []
            return

        pos = 0
        length = len(container)
        ret_part = [-1] * length
        while pos >= 0:
            ret_part[pos] += 2

            if ret_part[pos] > container[pos]*2 or (pos != 0 and ret_part[pos] > ret_part[pos - 1]):
                ret_part[pos] = -1
                pos -= 1
            else:
                pos += 1

            if pos == length:
                yield [QQ(n) / QQ(2) for n in ret_part]
                pos -= 1

    def to_virtual(self, rc):
        """
        Convert ``rc`` into a rigged configuration in the virtual crystal.

        INPUT:

        - ``rc`` -- a rigged configuration element

        EXAMPLES::

            sage: RC = RiggedConfigurations(CartanType(['A',4,2]).dual(), [[2,2]])
            sage: elt = RC(partition_list=[[1],[1]]); elt
            <BLANKLINE>
            -1[ ]-1
            <BLANKLINE>
            1[ ]1
            <BLANKLINE>
            sage: velt = RC.to_virtual(elt); velt
            <BLANKLINE>
            -1[ ]-1
            <BLANKLINE>
            2[ ]2
            <BLANKLINE>
            -1[ ]-1
            <BLANKLINE>
            sage: velt.parent()
            Rigged configurations of type ['A', 3, 1] and factor(s) ((2, 2), (2, 2))
        """
        gammatilde = list(self._folded_ct.scaling_factors())
        gammatilde[-1] = 2
        sigma = self._folded_ct.folding_orbit()
        n = len(self.virtual._rc_index)
        partitions = [None] * n
        for a,rp in enumerate(rc):
            g = gammatilde[a+1]
            for i in sigma[a+1]:
                partitions[i-1] = RiggedPartition(list(rp._list),
                                                  [rig_val*g for rig_val in rp.rigging])
        return self.virtual.element_class(self.virtual, partitions)

    def from_virtual(self, vrc):
        """
        Convert ``vrc`` in the virtual crystal into a rigged configuration of
        the original Cartan type.

        INPUT:

        - ``vrc`` -- a virtual rigged configuration element

        EXAMPLES::

            sage: RC = RiggedConfigurations(CartanType(['A',4,2]).dual(), [[2,2]])
            sage: elt = RC(partition_list=[[1],[1]])
            sage: velt = RC.to_virtual(elt)
            sage: ret = RC.from_virtual(velt); ret
            <BLANKLINE>
            -1[ ]-1
            <BLANKLINE>
            1[ ]1
            <BLANKLINE>
            sage: ret == elt
            True
        """
        gammatilde = list(self._folded_ct.scaling_factors())
        gammatilde[-1] = QQ(2)
        sigma = self._folded_ct.folding_orbit()
        n = len(self._rc_index)
        partitions = [None] * n
        # +/- 1 for indexing
        for a in range(n):
            rp = vrc[sigma[a+1][0] - 1]
            g = gammatilde[a+1]
            partitions[a] = RiggedPartition(list(rp._list),
                                            [rig_val/g for rig_val in rp.rigging])
        return self.element_class(self, partitions)

    Element = KRRCTypeA2DualElement



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.graphs sage.modules
r"""
Highest weight crystals
"""

#*****************************************************************************
#       Copyright (C) 2009   Anne Schilling <anne at math.ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#****************************************************************************

from sage.categories.classical_crystals import ClassicalCrystals
from sage.structure.parent import Parent
from sage.combinat.crystals.letters import CrystalOfLetters
from sage.combinat.crystals.tensor_product import TensorProductOfCrystals, \
    TensorProductOfRegularCrystalsElement

from sage.combinat.crystals.tensor_product import CrystalOfTableaux
from sage.combinat.crystals.alcove_path import CrystalOfAlcovePaths
from sage.combinat.crystals.littelmann_path import CrystalOfLSPaths
from sage.combinat.crystals.generalized_young_walls import CrystalOfGeneralizedYoungWalls
from sage.combinat.crystals.monomial_crystals import CrystalOfNakajimaMonomials
from sage.combinat.rigged_configurations.rc_crystal import CrystalOfRiggedConfigurations
from sage.rings.integer_ring import ZZ


def HighestWeightCrystal(dominant_weight, model=None):
    r"""
    Return the highest weight crystal of highest weight ``dominant_weight``
    of the given ``model``.

    INPUT:

    - ``dominant_weight`` -- a dominant weight
    - ``model`` -- (optional) if not specified, then we have the following
      default models:

      * types `A_n, B_n, C_n, D_n, G_2` - :class:`tableaux
        <sage.combinat.crystals.tensor_product.CrystalOfTableaux>`
      * types `E_{6,7}` - :class:`type E finite dimensional crystal
        <FiniteDimensionalHighestWeightCrystal_TypeE>`
      * all other types - :class:`LS paths
        <sage.combinat.crystals.littelmann_path.CrystalOfLSPaths>`

      otherwise can be one of the following:

      * ``'Tableaux'`` - :class:`KN tableaux
        <sage.combinat.crystals.tensor_product.CrystalOfTableaux>`
      * ``'TypeE'`` - :class:`type E finite dimensional crystal
        <FiniteDimensionalHighestWeightCrystal_TypeE>`
      * ``'NakajimaMonomials'`` - :class:`Nakajima monomials
        <sage.combinat.crystals.monomial_crystals.CrystalOfNakajimaMonomials>`
      * ``'LSPaths'`` - :class:`LS paths
        <sage.combinat.crystals.littelmann_path.CrystalOfLSPaths>`
      * ``'AlcovePaths'`` - :class:`alcove paths
        <sage.combinat.crystals.alcove_path.CrystalOfAlcovePaths>`
      * ``'GeneralizedYoungWalls'`` - :class:`generalized Young walls
        <sage.combinat.crystals.generalized_young_walls.CrystalOfGeneralizedYoungWalls>`
      * ``'RiggedConfigurations'`` - :class:`rigged configurations
        <sage.combinat.rigged_configurations.rc_crystal.CrystalOfRiggedConfigurations>`

    EXAMPLES::

        sage: La = RootSystem(['A',2]).weight_lattice().fundamental_weights()
        sage: wt = La[1] + La[2]
        sage: crystals.HighestWeight(wt)
        The crystal of tableaux of type ['A', 2] and shape(s) [[2, 1]]

        sage: La = RootSystem(['C',2]).weight_lattice().fundamental_weights()
        sage: wt = 5*La[1] + La[2]
        sage: crystals.HighestWeight(wt)
        The crystal of tableaux of type ['C', 2] and shape(s) [[6, 1]]

        sage: La = RootSystem(['B',2]).weight_lattice().fundamental_weights()
        sage: wt = La[1] + La[2]
        sage: crystals.HighestWeight(wt)
        The crystal of tableaux of type ['B', 2] and shape(s) [[3/2, 1/2]]

    Some type `E` examples::

        sage: C = CartanType(['E',6])
        sage: La = C.root_system().weight_lattice().fundamental_weights()
        sage: T = crystals.HighestWeight(La[1])
        sage: T.cardinality()
        27
        sage: T = crystals.HighestWeight(La[6])
        sage: T.cardinality()
        27
        sage: T = crystals.HighestWeight(La[2])
        sage: T.cardinality()
        78
        sage: T = crystals.HighestWeight(La[4])
        sage: T.cardinality()
        2925
        sage: T = crystals.HighestWeight(La[3])
        sage: T.cardinality()
        351
        sage: T = crystals.HighestWeight(La[5])
        sage: T.cardinality()
        351

        sage: C = CartanType(['E',7])
        sage: La = C.root_system().weight_lattice().fundamental_weights()
        sage: T = crystals.HighestWeight(La[1])
        sage: T.cardinality()
        133
        sage: T = crystals.HighestWeight(La[2])
        sage: T.cardinality()
        912
        sage: T = crystals.HighestWeight(La[3])
        sage: T.cardinality()
        8645
        sage: T = crystals.HighestWeight(La[4])
        sage: T.cardinality()
        365750
        sage: T = crystals.HighestWeight(La[5])
        sage: T.cardinality()
        27664
        sage: T = crystals.HighestWeight(La[6])
        sage: T.cardinality()
        1539
        sage: T = crystals.HighestWeight(La[7])
        sage: T.cardinality()
        56

    An example with an affine type::

        sage: C = CartanType(['C',2,1])
        sage: La = C.root_system().weight_lattice().fundamental_weights()
        sage: T = crystals.HighestWeight(La[1])
        sage: sorted(T.subcrystal(max_depth=3), key=str)
        [(-Lambda[0] + 3*Lambda[1] - Lambda[2] - delta,),
         (-Lambda[0] + Lambda[1] + Lambda[2] - delta,),
         (-Lambda[1] + 2*Lambda[2] - delta,),
         (2*Lambda[0] - Lambda[1],),
         (Lambda[0] + Lambda[1] - Lambda[2],),
         (Lambda[0] - Lambda[1] + Lambda[2],),
         (Lambda[1],)]

    Using the various models::

        sage: La = RootSystem(['F',4]).weight_lattice().fundamental_weights()
        sage: wt = La[1] + La[4]
        sage: crystals.HighestWeight(wt)
        The crystal of LS paths of type ['F', 4] and weight Lambda[1] + Lambda[4]
        sage: crystals.HighestWeight(wt, model='NakajimaMonomials')
        Highest weight crystal of modified Nakajima monomials of
         Cartan type ['F', 4] and highest weight Lambda[1] + Lambda[4]
        sage: crystals.HighestWeight(wt, model='AlcovePaths')
        Highest weight crystal of alcove paths of type ['F', 4] and weight Lambda[1] + Lambda[4]
        sage: crystals.HighestWeight(wt, model='RiggedConfigurations')
        Crystal of rigged configurations of type ['F', 4] and weight Lambda[1] + Lambda[4]
        sage: La = RootSystem(['A',3,1]).weight_lattice().fundamental_weights()
        sage: wt = La[0] + La[2]
        sage: crystals.HighestWeight(wt, model='GeneralizedYoungWalls')
        Highest weight crystal of generalized Young walls of
         Cartan type ['A', 3, 1] and highest weight Lambda[0] + Lambda[2]

    TESTS:

    Check that the correct crystal is constructed for the fundamental weights::

        sage: for ct in CartanType.samples(finite=True, crystallographic=True):  # long time
        ....:     L = ct.root_system().weight_lattice()
        ....:     La = L.fundamental_weights()
        ....:     for model in ['Tableaux', 'NakajimaMonomials', 'AlcovePaths', 'RiggedConfigurations']:
        ....:         if model == 'Tableaux' and ct.type() in ["E", "F"]:
        ....:             continue
        ....:         for wt in La:
        ....:             C = crystals.HighestWeight(wt, model=model)
        ....:             assert L.weyl_dimension(wt) == C.cardinality(), "wrong cardinality in %s, weight %s" % (ct, wt)
        ....:             assert C.highest_weight_vector().weight() == wt, "wrong weight in %s, weight %s" % (ct, wt)

    Same thing for weights constructed from the simple roots::

        sage: for ct in CartanType.samples(finite=True, crystallographic=True):
        ....:     L = ct.root_system().root_space()
        ....:     La = L.fundamental_weights_from_simple_roots()
        ....:     for model in ['Tableaux', 'NakajimaMonomials', 'AlcovePaths', 'RiggedConfigurations']:
        ....:         if model == 'Tableaux' and ct.type() in ["E", "F"]:
        ....:             continue
        ....:         for wt in La:
        ....:             C1 = crystals.HighestWeight(wt.to_ambient().to_weight_space(ZZ), model=model)
        ....:             C2 = crystals.HighestWeight(wt, model=model)
        ....:             assert C1 == C2
    """
    cartan_type = dominant_weight.parent().cartan_type()
    if model is None:
        if cartan_type.is_finite():
            if cartan_type.type() == 'E':
                model = 'TypeE'
            elif cartan_type.type() in ['A','B','C','D','G']:
                model = 'Tableaux'
            else:
                model = 'LSPaths'
        else:
            model = 'LSPaths'

    # Make sure dominant_weight in the weight space
    if cartan_type.is_finite():
        dominant_weight = dominant_weight.to_ambient().to_weight_space(ZZ)

    if model == 'Tableaux':
        # we rely on the specific choice of positive roots here
        # except in type G_2, the fundamental weights are realized by
        # vectors with weakly decreasing nonnegative integer (or in
        # type B_n and D_n, half-integer) entries
        sh = dominant_weight.to_ambient().to_vector()
        if cartan_type.type() == "G":
            sh = (-sh)[2:0:-1]
        return CrystalOfTableaux(cartan_type, shape=sh)

    if model == 'TypeE':
        if not cartan_type.is_finite() or cartan_type.type() != 'E':
            raise ValueError("only for finite type E")
        if cartan_type.rank() == 6:
            return FiniteDimensionalHighestWeightCrystal_TypeE6(dominant_weight)
        elif cartan_type.rank() == 7:
            return FiniteDimensionalHighestWeightCrystal_TypeE7(dominant_weight)
        raise NotImplementedError

    if model == 'NakajimaMonomials':
        return CrystalOfNakajimaMonomials(cartan_type, dominant_weight)

    if model == 'LSPaths':
        # Make sure it's in the (extended) weight space
        if cartan_type.is_affine():
            P = dominant_weight.parent().root_system.weight_space(extended=True)
        else:
            P = dominant_weight.parent().root_system.weight_space()
        wt = P.sum_of_terms((i, c) for i,c in dominant_weight)
        return CrystalOfLSPaths(wt)

    if model == 'AlcovePaths':
        return CrystalOfAlcovePaths(dominant_weight, highest_weight_crystal=True)

    if model == 'GeneralizedYoungWalls':
        if not cartan_type.is_affine():
            raise ValueError("only for affine types")
        if cartan_type.type() != 'A':
            raise NotImplementedError("only for affine type A")
        # Make sure it's in the weight lattice
        P = dominant_weight.parent().root_system.weight_lattice(extended=True)
        wt = P.sum_of_terms((i, c) for i,c in dominant_weight)
        return CrystalOfGeneralizedYoungWalls(cartan_type.rank()-1, wt)

    if model == 'RiggedConfigurations':
        return CrystalOfRiggedConfigurations(cartan_type, dominant_weight)

    raise ValueError("invalid model")


class FiniteDimensionalHighestWeightCrystal_TypeE(TensorProductOfCrystals):
    """
    Commonalities for all finite dimensional type `E` highest weight crystals.

    Subclasses should setup an attribute column_crystal in their
    ``__init__`` method before calling the ``__init__`` method of this class.
    """

    def __init__(self, dominant_weight):
        """
        EXAMPLES::

            sage: C = CartanType(['E',6])
            sage: La = C.root_system().weight_lattice().fundamental_weights()
            sage: T = crystals.HighestWeight(2*La[2])
            sage: T.cartan_type()
            ['E', 6]
            sage: T.module_generators
            [[[(2, -1), (1,)], [(2, -1), (1,)]]]
            sage: T.cardinality()
            2430
            sage: T = crystals.HighestWeight(La[2])
            sage: T.cardinality()
            78
        """
        self._cartan_type = dominant_weight.parent().cartan_type()
        self._highest_weight = dominant_weight
        assert dominant_weight.is_dominant()
        self.rename()
        Parent.__init__(self, category=ClassicalCrystals())
        self.module_generators = [self.module_generator()]

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: C = CartanType(['E',6])
            sage: La = C.root_system().weight_lattice().fundamental_weights()
            sage: crystals.HighestWeight(2*La[2])
            Finite dimensional highest weight crystal of type ['E', 6] and highest weight 2*Lambda[2]
        """
        return "Finite dimensional highest weight crystal of type {} and highest weight {}".format(
                self._cartan_type, self._highest_weight)

    Element = TensorProductOfRegularCrystalsElement

    def module_generator(self):
        """
        Yield the module generator (or highest weight element) of the classical
        crystal of given dominant weight in ``self``.

        EXAMPLES::

            sage: C=CartanType(['E',6])
            sage: La=C.root_system().weight_lattice().fundamental_weights()
            sage: T = crystals.HighestWeight(La[2])
            sage: T.module_generator()
            [[(2, -1), (1,)]]
            sage: T = crystals.HighestWeight(0*La[2])
            sage: T.module_generator()
            []

            sage: C=CartanType(['E',7])
            sage: La=C.root_system().weight_lattice().fundamental_weights()
            sage: T = crystals.HighestWeight(La[1])
            sage: T.module_generator()
            [[(-7, 1), (7,)]]
        """
        dominant_weight = self._highest_weight
        tensor = sum(( [self.column_crystal[i]]*dominant_weight.coefficient(i) for i in dominant_weight.support()), [])
        return self._element_constructor_(*[B.module_generators[0] for B in tensor])


class FiniteDimensionalHighestWeightCrystal_TypeE6(FiniteDimensionalHighestWeightCrystal_TypeE):
    r"""
    Class of finite dimensional highest weight crystals of type `E_6`.

    EXAMPLES::

        sage: C=CartanType(['E',6])
        sage: La=C.root_system().weight_lattice().fundamental_weights()
        sage: T = crystals.HighestWeight(La[2]); T
        Finite dimensional highest weight crystal of type ['E', 6] and highest weight Lambda[2]
        sage: B1 = T.column_crystal[1]; B1
        The crystal of letters for type ['E', 6]
        sage: B6 = T.column_crystal[6]; B6
        The crystal of letters for type ['E', 6] (dual)
        sage: t = T(B6([-1]),B1([-1,3])); t
        [(-1,), (-1, 3)]
        sage: [t.epsilon(i) for i in T.index_set()]
        [2, 0, 0, 0, 0, 0]
        sage: [t.phi(i) for i in T.index_set()]
        [0, 0, 1, 0, 0, 0]
        sage: TestSuite(t).run()
    """

    def __init__(self, dominant_weight):
        """
        EXAMPLES::

            sage: C=CartanType(['E',6])
            sage: La=C.root_system().weight_lattice().fundamental_weights()
            sage: p2=2*La[2]
            sage: p1=La[2]
            sage: p0=0*La[2]
            sage: T = crystals.HighestWeight(0*La[2])
            sage: T.cardinality()
            1
            sage: T = crystals.HighestWeight(La[2])
            sage: T.cardinality()
            78
            sage: T = crystals.HighestWeight(2*La[2])
            sage: T.cardinality()
            2430
        """
        B1 = CrystalOfLetters(['E',6])
        B6 = CrystalOfLetters(['E',6], dual=True)
        self.column_crystal = {1 : B1, 6 : B6,
                               4 : TensorProductOfCrystals(B1,B1,B1,generators=[[B1([-3,4]),B1([-1,3]),B1([1])]]),
                               3 : TensorProductOfCrystals(B1,B1,generators=[[B1([-1,3]),B1([1])]]),
                               5 : TensorProductOfCrystals(B6,B6,generators=[[B6([5,-6]),B6([6])]]),
                               2 : TensorProductOfCrystals(B6,B1,generators=[[B6([2,-1]),B1([1])]])}
        FiniteDimensionalHighestWeightCrystal_TypeE.__init__(self, dominant_weight)


class FiniteDimensionalHighestWeightCrystal_TypeE7(FiniteDimensionalHighestWeightCrystal_TypeE):
    r"""
    Class of finite dimensional highest weight crystals of type `E_7`.

    EXAMPLES::

        sage: C=CartanType(['E',7])
        sage: La=C.root_system().weight_lattice().fundamental_weights()
        sage: T = crystals.HighestWeight(La[1])
        sage: T.cardinality()
        133
        sage: B7 = T.column_crystal[7]; B7
        The crystal of letters for type ['E', 7]
        sage: t = T(B7([-5, 6]), B7([-2, 3])); t
        [(-5, 6), (-2, 3)]
        sage: [t.epsilon(i) for i in T.index_set()]
        [0, 1, 0, 0, 1, 0, 0]
        sage: [t.phi(i) for i in T.index_set()]
        [0, 0, 1, 0, 0, 1, 0]
        sage: TestSuite(t).run()
    """

    def __init__(self, dominant_weight):
        """
        EXAMPLES::

            sage: C=CartanType(['E',7])
            sage: La=C.root_system().weight_lattice().fundamental_weights()
            sage: T = crystals.HighestWeight(0*La[1])
            sage: T.cardinality()
            1
            sage: T = crystals.HighestWeight(La[1])
            sage: T.cardinality()
            133
            sage: T = crystals.HighestWeight(2*La[1])
            sage: T.cardinality()
            7371
        """
        B = CrystalOfLetters(['E',7])
        self.column_crystal = {7 : B,
                               1 : TensorProductOfCrystals(B,B,generators=[[B([-7,1]),B([7])]]),
                               2 : TensorProductOfCrystals(B,B,B,generators=[[B([-1,2]),B([-7,1]),B([7])]]),
                               3 : TensorProductOfCrystals(B,B,B,B,generators=[[B([-2,3]),B([-1,2]),B([-7,1]),B([7])]]),
                               4 : TensorProductOfCrystals(B,B,B,B,generators=[[B([-5,4]),B([-6,5]),B([-7,6]),B([7])]]),
                               5 : TensorProductOfCrystals(B,B,B,generators=[[B([-6,5]),B([-7,6]),B([7])]]),
                               6 : TensorProductOfCrystals(B,B,generators=[[B([-7,6]),B([7])]])}
        FiniteDimensionalHighestWeightCrystal_TypeE.__init__(self, dominant_weight)
